//************************************************************************* 
//* --------------------
//* IceRayTrace
//* --------------------
//* (Description)
//*  Toy Monte-Carlo for photon propagation in Ice
//*  Main reference document is Ped's Ph-D thesis
//*  "Muon energy reconstruction in the Antarctic Muon And Neutrino Detector Array"
//* (Requires)
//* (Provides)
//* (Update Record)
//*	2004/04/15  K.Hoshina	Original version.
//*************************************************************************

#include "raytrace/IceRayTrace.h"
#include "raytrace/PhotonBuf.h"
#include "raytrace/RayTracePhoton.h"
#include "romeo/kern/RMOUnit.h"
#include "TRandom.h"
#include "TMath.h"
#include <iostream>

ClassImp(IceRayTrace)

//=====================================================================
//* constructor -------------------------------------------------------

IceRayTrace::IceRayTrace(const char *name, const char *title)
       			//: RayTraceUtil(), RMOVModule(name,title),
       			: RMOVModule(name,title), RayTraceUtil(), 
                          fOutFile(0), fHdNdL(0)
{
   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTrace:constructor is called" << endl;
   }

   fOffsetShellRadius = 100. * kMETER;
   fRadiusStep        = 5. * kMETER;    
   fNshells           = 11;                 // 100m ~ 150m 
   fOutFileName       = "raytrace.root";
   fStoreHistory      = kFALSE;
}

//=====================================================================
//* destructor --------------------------------------------------------

IceRayTrace::~IceRayTrace()
{

   if (fOutFile)  delete fOutFile;
   if (fHdNdL)    delete fHdNdL;
   for (UInt_t i=0; i<fWLSpect_r.size(); i++) {
      if (fWLSpect_r[i]) delete fWLSpect_r[i];
   }

   // Are you sure you have set owner recursively?
   fShellBuffers.SetOwner();
   if (fShellBuffers.GetEntries()) fShellBuffers.RMOVAttBuffer::Delete();
}

//=====================================================================
//* Configure ---------------------------------------------------------

void IceRayTrace::Configure()
{
//
// Configure
// Initializing following items:
// - output rootfile
// - Create Histograms and output buffers
//

   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTrace::Configure is called!" << endl;
   }

   // 
   // create output rootfile
   //

   // save the original directory
   TDirectory *old_dir = gDirectory;

   // open file
   fOutFile = TFile::Open(fOutFileName, "recreate");
   fOutFile->cd("/");
   fOutFile->mkdir("functions", "functions");
   fOutFile->cd("/functions");

   //
   // make histograms and output buffers
   //

   fOutFile->cd("/");
   fOutFile->mkdir("hists", "hists");
   fOutFile->mkdir("photons", "photons");
   fOutFile->cd("/hists");

   Double_t lmin = GetLambdaMin();
   Double_t lmax = GetLambdaMax();
   Int_t nxbins = (Int_t)(lmax - lmin);
   fHdNdL       = new TH1D ("fHdNdL", "fHdNdL", nxbins, lmin, lmax); 
   fHdNdL->SetStats(1);

   for (Int_t i=0; i<fNshells; i++) {
      
      Int_t curR = fOffsetShellRadius + (Int_t)((i) * fRadiusStep);

      TString funcname("WLSpecR");
      funcname += curR;
      funcname += "m";
      fWLSpect_r.push_back(CreateCherenkovSpectrumAtDistFunc(curR)); 

      TString bufname("PhotonBuf");
      bufname += curR;
      bufname += "m";
      PhotonBuf *buf = new PhotonBuf(bufname);
      cerr << "Photon buf created! name " << bufname << endl;
      buf->SetDistance(curR);
      // You are the owner! set owner now.
      buf->SetOwner();
      fShellBuffers.Add(buf);
   }
   // don't forget to set owner to buffers
   fShellBuffers.SetOwner();
   fTmpHistoryBuffer.SetOwner();

   // back to the original directory
   old_dir->cd();
}

//=====================================================================
//* Process -----------------------------------------------------------

void IceRayTrace::Process()
{
//
// Process carries on following procedures.
// 
// 1) Choose a wavelength
// 2) Inject a photon from origin (0., 0., 0.) with a direction of 
//    (1, 0, 0). Note that the primary geometric-parameters are fixed.
//    Users must consider local-to-global coordinate transformation.
// 3) Stepping the photon as defined inside Stepping() function.
//    The step is stored RMOOpticalPhoton as a history of the step.
//    if the photon-path acrosses one of the pre-defined geometrical 
//    spheres (shells), save the crossing position and the ongoing 
//    direction into a RayTracePhoton object with its stepping history. 
//    Save the photon into buffer for that shell.
//    Repeat the Stepping until the photon is absorbed by ice or the 
//    accumulated traveling length is over 3 times of largest shell.
//

   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTrace::Process called" << endl;
   }

   // 0) initialization
   Bool_t isfilled[fNshells];
   for (Int_t i=0; i<fNshells; i++) {
      isfilled[i] = kFALSE;
   }

   // 1) Choose a wavelength
   
   //Double_t lambda = fWLSpectrumFunc->GetRandom(fLambdaMin, fLambdaMax);
   Double_t lambda = 337.0 * kNANO_METER;  // [m],  flasher wave length

   fHdNdL->Fill(lambda);

   //2) Inject a photon
   
   TVector3 firstdir(1, 0, 0);
   TVector3 oldpos(0, 0, 0);
   TVector3 olddir(firstdir);
   TVector3 newpos(0, 0, 0);
   TVector3 newdir(0, 0, 0);
   Double_t scatcostheta = 0;
   Double_t accumulatetime = 0.0;
   Bool_t   isabsorbed = kFALSE;

   TObjArray history;  // history buffer for this photon. its NOT an owner.

   Double_t currentdist     = newpos.Mag();
   Double_t accumulatedist  = newpos.Mag();

   Int_t curStepNo = 0;
   
   //
   // 3) stepping loop start
   //
   const TF1 & diffCrossSection = GetDiffCrossSectionFunc();
   while (accumulatedist < 3 * (fOffsetShellRadius + (fNshells + 1) * fRadiusStep)) { 

      // excute one step //
      // olddir and newdir are unit vectors //

      Double_t stepping_time = Stepping(lambda, oldpos, olddir, 
                                        diffCrossSection,
                                        newpos, newdir, scatcostheta, isabsorbed); 

      currentdist = newpos.Mag(); 
      Double_t currentsteplen = (newpos - oldpos).Mag();
      accumulatedist += currentsteplen;
      accumulatetime += stepping_time;

      if (GetVerboseLevel() == kDEBUG) {
         cerr << "*--- new step ---*" << endl; 
         cerr << "newpos, newdir, olddir, steplen, dist = (" 
              << newpos.X() << " " << newpos.Y() << " "
              << newpos.Z() << ") (" 
              << newdir.X() << " " << newdir.Y() << " "
              << newdir.Z() << ") (" 
              << olddir.X() << " " << olddir.Y() << " "
              << olddir.Z() << ") " 
              << (newpos - oldpos).Mag() << " " 
              << currentdist << endl;
      }
      
      // find intersection with shells and fill photons

      Int_t i     = 0;
      TIter iter = fShellBuffers.GetIter();
      PhotonBuf *buf;
      while ((Int_t)(buf = (PhotonBuf *)iter.Next())) {

         //Double_t curshell = fOffsetShellRadius + i * fRadiusStep; 
         Double_t curshell = buf->GetDistance();
         Double_t olddist  = oldpos.Mag();
         Double_t newdist  = newpos.Mag();
         Int_t    nfoundx  = 0;
         TVector3 x[2];

         //
         //To save CPU time, calculate special cases first
         //
         if (olddist < curshell && newdist < curshell) {
            // it is impossible to intersect with current shell...
            nfoundx = 0; 

         } else if (olddist > curshell && (oldpos * olddir) > 0) {
            // photon fly away from current shell
            nfoundx = 0; 

         } else if (newdist == curshell) {
            // only one intersection 
            x[0] = newpos;
            nfoundx = 1;

         } else {
            // find intersection within the step range
            nfoundx = FindIntersection(oldpos, olddir, newpos, curshell, x);
         }

         for (Int_t k=0; k<nfoundx; k++) {

            RayTracePhoton *ph = new RayTracePhoton(x[k], olddir);
            ph->SetWaveLength(lambda);    // [m]
            ph->SetTime(accumulatetime);  // [sec]
            ph->SetCosTheta(scatcostheta);   // for debugging purpose

            if (fStoreHistory) {
               // copy history to the RayTracePhoton
               ph->SetHistory(history);
            }

            if (GetVerboseLevel() == kDEBUG) {
               ph->DumpInfo();
            }

            buf->Add(ph); 
         }
         i++;

      } // shells loop end

      //
      // if photon is absorbed, do not continue the stepping loop
      //
      if (isabsorbed) {
         if (GetVerboseLevel() > kPRECISE) {
            cerr << "*** Photon is absorbed at pos, accumdist = "
                 << newpos.X() << " " << newpos.Y() << " "
                 << newpos.Z() << " " << accumulatedist << " ***"<< endl;
         }
         break;
      }

      //
      // update olddata and counter incriment //
      //
      oldpos  = newpos;
      olddir  = newdir;

      curStepNo++;

      if (! curStepNo % 100) {
         cerr << curStepNo << " step finished!" << endl;
      }

      // store current history into the history buffer
      if (fStoreHistory) {
         RMOOpticalPhoton *histph = new RMOOpticalPhoton(newpos, newdir);
         histph->SetWaveLength(lambda);    // [m]
         histph->SetTime(accumulatetime);  // [sec]
         histph->SetCosTheta(scatcostheta);   // for debugging purpose
         if (GetVerboseLevel() == kDEBUG) {
            cerr << "new histph " << histph << " is created!" << endl;   
         }
         history.Add(histph);
         fTmpHistoryBuffer.Add(histph);
      }

   } // stepping loop end

}

//=====================================================================
//* Finish ------------------------------------------------------------

void IceRayTrace::Finish()
{
   fOutFile->cd("/photons");
   fOutFile->mkdir("history");
   TIter next = fShellBuffers.GetIter();
   while (PhotonBuf *buf = (PhotonBuf *)next.Next()) {
      fOutFile->cd("/photons");
      buf->Write();
   }
   fShellBuffers.Write();
   
   fOutFile->cd("/");
   fOutFile->Write();
   fOutFile->Close();

   // now we can delete tmp buffers...
   fTmpHistoryBuffer.Delete();
}

//=====================================================================
//* Draw ------------------------------------------------------------

void IceRayTrace::Draw()
{
   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTrace::Draw called" << endl;
   }

   RayTraceUtil::Draw();
}


//================================================================
// function FindIntersection
//================================================================

Int_t IceRayTrace::FindIntersection(const TVector3 &oldpos, 
                                 const TVector3 &olddir,
                                 const TVector3 &newpos,
                                 Double_t curshell,
                                 TVector3 intersection[]) 
{
   //
   // find intersection of current shell and current step
   // Particle starts from oldpos with olddir and reache to newpos
   // Equations:
   // x = oldpos + t * olddir;
   // |x|^2 = curshell^2;
   // a*t^2 + 2*b*t + c = 0;
   //   where 
   //   a = |olddir|^2
   //   b = oldpos * olddir
   //   c = |oldpos|^2 - curshell^2
   //

   Double_t a = olddir.Mag2();
   Double_t b = oldpos * olddir;
   Double_t c = oldpos.Mag2() - curshell * curshell;

   // Is there any intersection?
   if ((b*b - a*c) < 0) {
      // no intersection!
      return 0;
   }

   Double_t d    = TMath::Sqrt(b*b - a*c);
   Double_t tpos = (-b + d) / a;
   Double_t tneg = (-b - d) / a;
   Double_t t[2];
   Int_t nfound  = 0;


   // 
   // Find intersection. 
   // Note that parameter t must be within a
   // range tmin < t <= tmax
   // (oldpos shouldn't be counted as a intersection of current step)
   //

   Double_t tmin = 0.;  // at oldpos
   Double_t tmax = (newpos - oldpos).Mag();  // at newpos

   if (d == 0.) { 
      if (tpos > tmin && tpos <= tmax) {  // tpos == tneg
         nfound = 1;
         t[0] = tpos;
      }
   } else { // d > 0
      if (tneg > tmin && tneg <= tmax) {
         nfound = 1;
         t[0] = tneg;
      }
      if (tpos > tmin && tpos <= tmax) {
         if (nfound) {
            t[1] = tpos;
         } else {
            t[0] = tpos;
         }
         nfound ++;
      }
   }


   for (Int_t i=0; i<nfound; i++) {
      intersection[i] = (oldpos + t[i]*olddir);
   }

   return nfound;

}

