// $Id: RMOVModule.cxx,v 1.1.1.1 2005/06/23 11:43:48 hoshina Exp $
//*************************************************************************
//* --------------------
//* RMOVModule
//* --------------------
//* (Description)
//*	Abstract base class for RMO modules.
//* (Requires)
//* (Provides)
//* (Update Record)
//*	2003/12/22  K.Hoshina	Original version.
//*************************************************************************

#include "romeo/kern/RMOVModule.h"
#include "romeo/kern/RMOVAttBuffer.h"

ClassImp(RMOVModule)

using namespace std;

EState      RMOVModule::fgState    = kALIVE;
RMOVModule* RMOVModule::fgLastTask = 0;

//=====================================================================
//* constructor -------------------------------------------------------

RMOVModule::RMOVModule(const char *name, const char *title)
       			: TTask(name,title), RMOAttVerbose(),
                          fHasExecutedProcess(kFALSE)
{
//*-*-*-*-*-*-*-*-*-*Constructor for module*-*-*-*-*-*-*-*-*-*-*
//                  =======================

   VerbosePrint(TString("RMOVModule::constructor of ") + GetName() 
                + "is called", kDEBUG);
}

//=====================================================================
//* destructor --------------------------------------------------------

RMOVModule::~RMOVModule()
{
//*-*-*-*-*-*-*-*-*-*Destructor for module*-*-*-*-*-*-*-*-*-*-*
//                   =====================
}

//=====================================================================
//* Add ---------------------------------------------------------------

void RMOVModule::Add(RMOVModule *module)
{
//*-*-*-*-*-*-*-*-*-*-*-*Add sub module*-*-*-*-*-*-*-*-*-*-*-*
//                       ==============
//  Use this Add method instead of TTask::Add, in order to 
//  set mother module at once.

    VerbosePrint(TString("RMOVModule::Add: ") + module->GetName()
                         + " is added to " + GetName(), kDEBUG);
    TTask::Add(module);
    module->SetMother(this);
}

void RMOVModule::Add(TTask *module)
{
//  Use RMOVModule::Add(RMOVModule *module) 
//  instead of it.
    
    Fatal("Add", "Use Other Add() function.");
    abort();
}

//=====================================================================
//* Exec --------------------------------------------------------------

void RMOVModule::Exec(Option_t *opt)
{
//*-*-*-*-*-*-*-*-*-*-*-*Execute task*-*-*-*-*-*-*-*-*-*-*-*-*
//                       ============
//  The Exec function is called by ExecuteTask() or ExecuteTasks() 
//  automatically. 
//  Options must be one of pure-virtual method. 
//  Current options are:
//     Configure : for configuration all modules
//     Process   : main process 
//     Finish    : executed at the end of run 
//     Draw      : Draw histgrams 
//  
//*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

   if (fgBeginTask == this) fgState = kALIVE;

   TString option(opt);

   if (fgState != kALIVE) {
      if (option == "Process") fHasExecutedProcess = kFALSE;
      return;
   }

   if (option == "Configure") {
      VerbosePrint(TString("RMOVModule::Exec: Configure of ")
                   + GetName() + " is called!", kFUNCTION_CALLED);
      Configure();
      fHasExecutedProcess = kFALSE;   

   } else if (option == "Process") {
      VerbosePrint(TString("RMOVModule::Exec: Process of ")
                   + GetName() + " is called!", kFUNCTION_CALLED);
      Process();
      fHasExecutedProcess = kTRUE;   

   } else if (option == "Finish") {  
      VerbosePrint(TString("RMOVModule::Exec: Finish of ")
                   + GetName() + " is called!", kFUNCTION_CALLED);
      Finish();

   } else if (option == "Draw") {  
      VerbosePrint(TString("RMOVModule::Exec: Draw of ")
                   + GetName() + " is called!", kFUNCTION_CALLED);
      Draw();

   } else if (option == "ClearBuffer") {  
      VerbosePrint(TString("RMOVModule::Exec: ClearBuffer of ")
                   + GetName() + " is called!", kFUNCTION_CALLED);
      ClearBuffer();

   } else if (option == "ClearEvent") {  
      VerbosePrint(TString("RMOVModule::Exec: ClearEvent of ")
                   + GetName() + " is called!", kFUNCTION_CALLED);
      ClearEvent();
   }


   if (fgState != kALIVE) {
      switch (fgState) { 
         case kVANISHED:
            VerbosePrint(TString("RMOVModule::Exec:  The particle vanished at ")
                         + GetName(), kPRECISE);
            break;
         case kABSORBED:
            VerbosePrint(TString("RMOVModule::Exec:  The particle absorbed at ")
                         + GetName(), kPRECISE);
            break;
         case kABORTED:
            VerbosePrint(TString("RMOVModule::Exec:  The program is aborted at ")
                         + GetName(), kPRECISE);
            break;
         default:
            VerbosePrint(TString("RMOVModule::Exec:  Wrong Code! something strange happen ")
                         + GetName(), kPRECISE);
            break;
      }
   }
   fgLastTask = this;
}

//=====================================================================
//* GetImputBuffer ----------------------------------------------------

const TObject& RMOVModule::GetImputBuffer()
{
//*-*-*-*-*-*-*-*-*-*-*-*Return buffer*-*-*-*-*-*-*-*-*-*-*-*-*
//                       =============
//  It returns reference to imput buffer associeted with forward task(RMOVModule).
//  If the module is first task, it returns buffer of mother task.
//     
//*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*

   RMOVModule *pre = (RMOVModule *)GetMother()->GetListOfTasks()->Before(this);
   if (!pre) pre = GetMother();
   TObject *buf = const_cast<TObject *>(&pre->GetBuffer());
   if (! dynamic_cast<RMOVAttBuffer *>(buf)) {
      Fatal("GetImputBuffer", "Buffer does not inherit RMOVAttBuf!");
   }
      
   return *buf;

}

//=====================================================================
//* ClearBuffer  ------------------------------------------------------

void RMOVModule::ClearBuffer()
{
//*-*-*-*-*-*-*-*-*-*-*-*Clear Buffer*-*-*-*-*-*-*-*-*-*-*-*-*
//                       ============
// Remove all object pointers from the buffer array.
// Does not delete the objects unless the buffer is the owner of objects
// (set via buf->SetOwner()).
//

   TObject *b = const_cast<TObject *>(&GetBuffer());
   RMOVAttBuffer *buf = 0;
   if (b) buf = dynamic_cast<RMOVAttBuffer *>(b);
   if (buf) { 
      buf->Clear();
   } else {
      VerbosePrint(TString("RMOVModule::ClearBuffer: Dinamic cast failed")
                  + GetName(), kDEBUG);
   }
}

//=====================================================================
//* ClearEvent  -------------------------------------------------------

void RMOVModule::ClearEvent()
{
//*-*-*-*-*-*-*-*-*-*-*-*Clear Event*-*-*-*-*-*-*-*-*-*-*-*-*
//                       ===========
// Clear all flags and buffer.

   fgState = kALIVE;
   fgLastTask = 0;
   fHasExecutedProcess = kFALSE; 
   ClearBuffer();

}
