#ifndef __IceRayTrace__H__
#define __IceRayTrace__H__

//*************************************************************************
//* ---------------------
//* IceRayTrace Class
//* ---------------------
//* (Description)
//*     A handy raytracer for BULK ice. 
//*     (uses ROOT and ROMEO/kern, easy to make compatible with IceTray)
//*     Photons are injected from origin with fixed direction (1, 0, 0)
//*     and caught by the layers of spherical shells which center is fixed
//*     at origin.
//*     Once a photon reached to any shell, it records the arrival position,
//*     direction, arrival time and wavelength inside a RayTracePhoton.
//*     These RayTracePhotons are stored inside PhotonBuf assigned to each
//*     shells, then written into TFile at the end of run.
//*     
//* (Requires)
//*     libRMOKern, RayTraceUtil
//* (Update Record)
//*	2004/04/15  K.Hoshina	Original version.
//*************************************************************************

#include "TVector3.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TF1.h"
#include "TRandom.h"
#include "TFile.h"
#include "TObjArray.h"
#include <iomanip>
#include <vector>

#include "raytrace/PhotonBuf.h"
#include "raytrace/RayTraceUtil.h"
#include "romeo/kern/RMOVModule.h"

using namespace std;

class IceRayTrace : public RMOVModule, public RayTraceUtil {
//class IceRayTrace : public RayTraceUtil, public RMOVModule {

public:

  IceRayTrace(const Char_t *name = "IceRayTrace", 
         const Char_t *title = "IceRayTrace module");
  virtual ~IceRayTrace();

  virtual void Configure();
  virtual void Process()  ;
  virtual void Finish()   ;
  virtual void Draw();
  virtual void DumpInfo(){}

  virtual const TObject &GetBuffer() const { return fShellBuffers; }

  inline void  SetNshells(Int_t n)       { fNshells    = n; }
  inline void  SetRadiusStep(Double_t x) { fRadiusStep = x; }
  inline void  SetOffsetShellRadius(Double_t x) { fOffsetShellRadius = x; }
  inline void  SetOutFileName(const TString &s) { fOutFileName = s; }
  inline void  SetStoreHistory(Bool_t b = kTRUE) { fStoreHistory = b; }

private:
  static Int_t    FindIntersection(const TVector3 &oldpos,
                                   const TVector3 &olddir,
                                   const TVector3 &newdir,
                                   Double_t curshell,
                                   TVector3 intersection[]);


private:

  Double_t  fOffsetShellRadius; // offset radius of shells
  Int_t     fNshells;         // Number of shells 
  Double_t  fRadiusStep;      // distance between a shell to the next shell [m]

  TFile    *fOutFile;         // TFile to store the results
  TString   fOutFileName;     // name of the TFile

  PhotonBuf fShellBuffers;    // array of buffers to keep arrival photons at i-th shell 
  TObjArray fTmpHistoryBuffer;// temporary stepping history buffer

  vector<TF1*> fWLSpect_r;    //! Function of Wave Length  
  TH1      *fHdNdL;           // dN/dLambda

  Bool_t    fStoreHistory;    // flag to store stepping history

  ClassDef(IceRayTrace, 1)       // Ray trace Monte-Carlo program

};

#endif

