//
// global constants
//
//const Int_t    kNAngBins = 180;  // zenith & azimuth nbins
const Int_t    kNAngBins = 18;
const Int_t    kNCosThetaBins = 201; // corresponds 1m after 100m propagation
//const Int_t    kNCosThetaBins = 41; // corresponds ~5m after 100m propagation
const Double_t kCosThetaBinSize = (2.0 / kNCosThetaBins); 
const Double_t kCosThetaBinEdge = (1.0 - kCosThetaBinSize);
const Double_t kPi = TMath::Pi();
const Double_t kRadToDeg = (180. / kPi);
const Char_t * kOutputHistSuf = "_norm";

//
// Weight (1 / 2*PI*R) for each costheta bin of input TH3D histogram 
// It adds/recreates new histgram into input .root file.
//
Int_t Normalize(const TString &inputfname)
{
   gSystem->Load("libPhysics.so");
   gSystem->Load("../lib/libRMOKernel.so");
   gSystem->Load("../lib/libraytrace.so"); 
   gStyle->SetOptStat(kTRUE);

   TFile *infile = TFile::Open(inputfname, "UPDATE");
   infile->cd("/");

   //
   // loop over all histograms
   // a histogram stores position/direction of arrival photons
   // trapped by the surface of a sphere-shell. 
   // each histgram has a name which indecates the radius of
   // its specific sphere [distance].
   //
   TList *keys = infile->GetListOfKeys();
   for (Int_t ibuf=0; ibuf<keys->GetSize(); ibuf++) {

      // get buffer name
      cerr << "loop " << ibuf << endl;
      TKey *key = keys->At(ibuf);
      char *hname = key->GetTitle();
      cerr << "hname = " << hname << endl;

      // use only input hists..
      if (strstr(hname, kOutputHistSuf)) continue;

      // get distance from the name...      
      strstream tmpstr;
      for (Int_t i=0; i<strlen(hname); i++) {
         if (isdigit(hname[i])) tmpstr << hname[i];
      }
      Double_t distance;
      tmpstr >> distance;
      cerr << "distance = " << distance << endl;

      // get an object associates with the key
      TH3D *hist = (TH3D*)key->ReadObj();

      // create new hist to store weighted hist
      TString n_hname(hname);
      n_hname += kOutputHistSuf;
      TH3D *n_hist = new TH3D(*hist);
      n_hist->SetNameTitle(n_hname, n_hname);

      //
      // loop over opening angle bins to calculate weight
      // axis x ... opening angle costheta (-1<x<1)
      // axis y ... injection zenith at DOM local coordinate (0<y<180 deg)
      // axis z ... injection azimuth at DOM local coordinate (0<z<360 deg)
      //
      Int_t nbinsx = hist->GetNbinsX();
      Int_t nbinsy = hist->GetNbinsY();
      Int_t nbinsz = hist->GetNbinsZ();
      Double_t widthx = 2. / nbinsx;
      Double_t widthy = 180. / nbinsy;
      Double_t widthz = 360. / nbinsz;
      for (Int_t i=1; i<nbinsx - 1; i++) { // weight of first/last bin is 1

         Double_t curcostheta = -1 + (i + 0.5) * widthx; // center of the bin
         Double_t cursintheta = sqrt((1 + curcostheta)*(1 - curcostheta));
         Double_t r_perp = distance * cursintheta;
         Double_t weight = 1 / (2. * kPi * r_perp);
         //if (weight > 1.0) weight = 1.0; // weight > 1 corresponds 30cm diameter circle...
         //Double_t weight = distance / r_perp;
         //Double_t weight = 1.0;
        
         //
         // loop over zenith/azimuth to apply the weight
         //
         for (Int_t j=0; j<nbinsy; j++) {
            Double_t curzenith = (j + 0.5) * widthy;

            for (Int_t k=0; k<nbinsz; k++) {
               Double_t curazimuth = (k + 0.5) * widthz;
               Int_t curbin = hist->FindBin(curcostheta, curzenith, curazimuth);
               if (distance == 100 && curcostheta > 0.5) {
                  cerr << "x, y, z, curbin = " 
                       << curcostheta << " " << curzenith << " " << curazimuth 
                       << " " << curbin << endl;
               }
               Double_t curcontent = hist->GetBinContent(curbin);
               curcontent *= weight;
               if (distance == 100 && curcostheta > 0.5) {
                  cerr << "dist, r_perp, weight = " << distance << " " 
                       << r_perp << " " << weight << endl; 
                  cerr << "curcontent = " << hist->GetBinContent(curbin) << endl;
                  cerr << "weighted content = " << curcontent << endl;
               }
               n_hist->SetBinContent(curbin, curcontent);
            }
         }
               
      } // xaxis loop end
      
   } // buffer loop end
   infile->Write("0", TObject::kOverwrite);

}

