//
// global constants
//
const Int_t    kNTimeBins       = 500;
const Int_t    kNZenithAngBins  = 18;
const Int_t    kNAzimuthAngBins = 36;
const Int_t    kNCosThetaBins = 201;
const Double_t kCosThetaBinSize = (2.0 / kNCosThetaBins); 
const Double_t kCosThetaBinEdge = (1.0 - kCosThetaBinSize);
const Double_t kPi = TMath::Pi();
const Double_t kRadToDeg = (180. / kPi);

// Plot arrival photon(distance, opening angle [costheta]) and 
// injection angle at DOM position (injection zenith and azimuth)
// into TH4D histgram.
// It creates one TH4D histgram per distance which x-axis corresponds
// opening costheta, y-axis corresponds injection zenith, and 
// z-axis corresponds injection azimuth.
//
// Arguments:
//  ifname  ... input filename (with suffix)
//  ofname  ... output filename (with suffix)
//  option  ... save file option
// If option = NEW or CREATE   create a new file and open it for writing,
//                             if the file already exists the file is
//                             not opened.
//           = RECREATE        create a new file, if the file already
//                             exists it will be overwritten.
//           = UPDATE          open an existing file for writing.
//                             if no file exists, it is created.
//
Int_t Plot4D(TString ifname, TString ofname, char* option)
{
   gSystem->Load("libPhysics.so");
   gSystem->Load("../lib/libRMOKernel.so");
   gSystem->Load("../lib/libraytrace.so"); 
   gSystem->Load("../lib/libutils.so"); 
   gStyle->SetOptStat("ne");

   //gDebug = 1;

   TFile *infile = TFile::Open(ifname, "read");

   // open output file
   TFile *outfile = TFile::Open(ofname, option);

   // get buffer
   infile->cd("/photons");
   PhotonBuf *buf = (PhotonBuf *)gROOT->FindObject("PhotonBuf");
   cerr << "bufname = " << buf->GetName() << endl;

   // define global unit vectors 
   TVector3 ex(1., 0., 0.);
   TVector3 ey(0., 1., 0.);
   TVector3 ez(0., 0., 1.);

   // define local unit vectors at DOM
   TVector3 l_ex = ex;     // anytime same as global unit x
   TVector3 l_ey = ey;     // depend on the DOM position
   TVector3 l_ez = ez;     // depend on the DOM position

   //
   // loop over all buffers 
   // a buffer stores position/direction of arrival photons
   // trapped by the surface of a sphere-shell.
   // each buffer has a name which indecates the radius of
   // its specific sphere [distance].
   //
   TIter iter = buf->GetIter();
   PhotonBuf *buftmp;
   while ((Int_t)(buftmp = (PhotonBuf *)iter.Next())) {

      Double_t dist = buftmp->GetDistance();
      TString bufname = buftmp->GetName();
      cerr << "buftmpname = " << bufname << endl;

      // make/find output directory and file
      TDirectory *input_dir = gDirectory;
      outfile->cd("/");
      TString histname(bufname);
      histname += "_hist";
      TH4D *angHist = 0;

      if (option == "UPDATE" || option ==  "update") {
         // find object
         angHist = (TH4D *)gROOT->FindObject(histname);

      } else {
         // create new hist
         TH4D *angHist = new TH4D(histname, histname, 
                               kNCosThetaBins, -1.0, 1.0, 
                               kNZenithAngBins, 0., 180.,  // nadia angle of injection dir
                               kNAzimuthAngBins, 0., 360., // azimuth angle of injection dir
                               kNTimeBins, 0., 1500e-9.);  // arrival time [sec]
         angHist->GetXaxis()->SetTitle("Opening Angle[costheta]");
         angHist->GetXaxis()->SetTitleOffset(1.5);
         angHist->GetYaxis()->SetTitle("Local injection zenith[deg]");
         angHist->GetYaxis()->SetTitleOffset(1.5);
         angHist->GetZaxis()->SetTitle("Local injection azimuth[deg]");
         angHist->GetZaxis()->SetTitleOffset(1.5);
         angHist->GetExtraAxis()->SetTitle("arrival time [s]");
         angHist->SetStats(kTRUE);
      }
 
      //
      // loop over all photons to store them into TH4D
      //
      TIter iter2 = buftmp->GetIter();
      RMOOpticalPhoton *ph;
      while ((Int_t)(ph = (RMOOpticalPhoton *)iter2.Next())) {

         Double_t time = ph->GetTime();         // in [sec]
         TVector3 pos  = ph->GetPosition();     // arrival position
         TVector3 dir  = -(ph->GetDirection()); // incoming direction
         Double_t posx = pos.X();
         Double_t costheta = posx / dist;       // cosine between x-axis and pos
         
	 //
	 //Set local coordinate at DOM position
         //if the -0.999 < costheta < 0.999 
         //(except first/last bins, 1m vertical length at DOM position)
	 //
         if (TMath::Abs(costheta) < kCosThetaBinEdge) {
            l_ez = (pos - TVector3(posx, 0, 0)).Unit();
            l_ey = l_ez.Cross(l_ex);
         }
         // calculate injection angle at local coordinate of DOM
         Double_t local_zenith_angle = l_ez.Angle(dir); 
         local_zenith_angle *= kRadToDeg;
         Double_t local_azimuth_angle = TMath::ATan2(l_ey * dir, l_ex * dir); // -PI < azimuth < PI
         if (local_azimuth_angle < 0) local_azimuth_angle += (2 * kPi); // 0 < azimuth < 2*PI
         local_azimuth_angle *= kRadToDeg;

         // fill histgram
         angHist->Fill(costheta, local_zenith_angle, local_azimuth_angle, time, 1);
              
      } // photon loop end

      input_dir->cd();  // return to imput directory

   } // buffer loop end

   infile->Close();

   //
   // save histgram into TFile
   //
   //outfile->Write("0", TObject::kWriteDelete);
   outfile->Write("0", TObject::kOverwrite);

   outfile->Close();

}


