//************************************************************************* 
//* --------------------
//* IceRayTraceSim
//* --------------------
//* (Description)
//*  Simulator for bulk ice ray-trace. If former module is not 
//*  the IceRayTrace, it uses the output root file of IceRayTrace
//* (Update Record)
//*	2005/12/15  K.Hoshina	Original version.
//*************************************************************************

#include "raytrace/IceRayTraceSim.h"
#include "raytrace/PhotonBuf.h"
#include "romeo/kern/RMOOpticalPhoton.h"
#include "romeo/kern/RMOUnit.h"
#include "TRandom.h"
#include "TMath.h"
#include <iostream>

ClassImp(IceRayTraceSim)

using namespace std;

//=====================================================================
//* constructor -------------------------------------------------------

IceRayTraceSim::IceRayTraceSim(const char *name, const char *title)
       			: RMOVModule(name,title), fInputRootFileName("raytrace.root")
{
   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTraceSim:constructor is called" << endl;
   }
}

//=====================================================================
//* destructor --------------------------------------------------------

IceRayTraceSim::~IceRayTraceSim()
{

}

//=====================================================================
//* Configure ---------------------------------------------------------

void IceRayTraceSim::Configure()
{
   //
   // Configure
   // Initializing following items:
   // - input rootfile

   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTraceSim::Configure is called!" << endl;
   }

   // 
   // read imput rootfile
   //
   if (gSystem->AccessPathName(fInputRootFileName) {
      // if the module doesn't have IceRayTrace module as a sub-module,
      // it doesn't work...
      TIter next(fTasks);
      TTask *task;
      Bool_t found = kFALSE;
      while((task=(TTask*)next())) {
         task->GetName() == "IceRayTrace" ? found = kTRUE : ;
      }

      if (!found) {
         Fatal("Configure", "Prepare output file from IceRayTrace or"
            " set IceRayTrace module as its sub-module!");
      }
      
   } else {
      // Found an input root file. Open it now.
      fInputFile = TFile::Open(fInputRootFileName, "read");
   }
}

//=====================================================================
//* Process -----------------------------------------------------------

void IceRayTraceSim::Process()
{
//
// The process carries on following procedures.
// 
// 1) Calculate transformation matrix from global to IceRayTrace local
// 2-1) If this module has any sub-module, execute it
// 2-2) Else, get a proper result from imput file

   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTraceSim::Process called" << endl;
   }
   
   // get input photons which is obaying global coordinate
   
   RMOPhotonBuf imputbuf = (PhotonBuf &)GetImputBuffer();
   TIter        iter     = imputbuf.GetIter();
   
   /* Create a new buffer. --------------- */
   // The name of buffer MUST BE IDENTICAL.
   // Compose buffer name from its module name.
   if (fOutputPhotonBuf) delete fOutputPhotonBuf;
   TString bufname(GetName());
   bufname += "Buf";
   fOutputPhotonBuf = new PhotonBuf(bufname);

   // Don't forget set the owner flag to the buffer
   // to avoid memory reak.
   fOutputPhotonBuf->SetOwner();

   // calculate transformation matrix
   // 
   RMOOpticalPhoton *iph;
   while ((Int_t)(iph = (RMOOpticalPhoton *)iter.Next())) {
      TVector3 global_pos = iph->GetPosition();
      TVector3 global_dir = iph->GetDirection();


   // save the original directory
   TDirectory *old_dir = gDirectory;

   // move to the calibration file directory

   fInputFile->cd("/photons"); 

   /* Get Input Data --------------------- */
   PhotonBuf *buf = (PhotonBuf *)gROOT->FindObject("PhotonBuf");
   cerr << "bufname = " << buf->GetName() << endl;



   //Double_t lambda = fWLSpectrumFunc->GetRandom(fLambdaMin, fLambdaMax);
   Double_t lambda = 337.0 * kNANO_METER;  // [m],  flasher wave length


   fHdNdL->Fill(lambda);
   Double_t dcsfpar[1];
   dcsfpar[0] = lambda;
   fDsDcosthetaFunc->SetParameters(dcsfpar);

   TVector3 firstdir(1, 0, 0);
   TVector3 grandolddir(firstdir);
   TVector3 oldpos(0, 0, 0);
   TVector3 olddir(firstdir);
   TVector3 newpos(0, 0, 0);
   TVector3 newdir(0, 0, 0);
   Bool_t   isabsorbed = kFALSE;

   Double_t currentdist     = newpos.Mag();
   Double_t accumulatedist  = newpos.Mag();

   fCurStepNo = 0;
   
   //
   // stepping loop start
   //


   while (accumulatedist < 3 * (fOffsetShellRadius + (fNshells + 1) * fRadiusStep)) { 

      // excute one step //
      // olddir and newdir are unit vectors //

      Stepping(lambda, oldpos, olddir, fDsDcosthetaFunc,
               newpos, newdir, isabsorbed); 

      currentdist = newpos.Mag(); 
      Double_t currentsteplen = (newpos - oldpos).Mag();
      accumulatedist += currentsteplen;

      if (GetVerboseLevel() == kDEBUG) {
         cerr << "currentpos, steplen, currentdist = " 
              << newpos.X() << " " << newpos.Y() << " "
              << newpos.Z() << " " 
              << (newpos - oldpos).Mag() << " " 
              << currentdist << endl;
      }
      
      if (isabsorbed) {

         if (GetVerboseLevel() == kPRECISE) {
            cerr << "Photon is absorbed! pos, accumdist = "
                 << newpos.X() << " " << newpos.Y() << " "
                 << newpos.Z() << " " << accumulatedist << endl;
         }

         break;

      }

      if (fCurStepNo < fNshells) {
         Double_t curcostheta  = GetCosTheta(newpos, olddir);
         Double_t curcosthetaz = GetCosTheta(firstdir, olddir);
         fHcostheta_i[fCurStepNo]->Fill(curcostheta);
         fHcostheta_zi[fCurStepNo]->Fill(curcosthetaz);
      }

      // fill photonarray //

      Int_t i     = 0;
      TIter iter = fShellBuffers.GetIter();
      PhotonBuf *buf;
      while ((Int_t)(buf = (PhotonBuf *)iter.Next())) {
         //Double_t curshell = fOffsetShellRadius + i * fRadiusStep; 
         Double_t curshell = buf->GetDistance();
         Double_t olddist  = oldpos.Mag();
         Double_t newdist  = newpos.Mag();
         Int_t    nfoundx  = 0;
         TVector3 x[2];

         //
         //To save CPU time, calculate special cases first
         //
         if (olddist < curshell && newdist < curshell) {
            // it is impossible to intersect with current shell...
            nfoundx = 0; 

         } else if (olddist > curshell && (oldpos * olddir) > 0) {
            // photon fly away from current shell
            nfoundx = 0; 

         } else if (newdist == curshell) {
            // only one intersection 
            x[0] = newpos;
            nfoundx = 1;

         } else {
            // find intersection within the step range
            nfoundx = FindIntersection(oldpos, olddir, newpos, curshell, x);
         }

         for (Int_t k=0; k<nfoundx; k++) {

            RMOOpticalPhoton *ph = new RMOOpticalPhoton(x[k], olddir);
            if (GetVerboseLevel() == kDEBUG) {
               cerr << "dist, pos, dir = " << curshell << " " 
                    << x[k].X() << " " << x[k].Y() << " " << x[k].Z() << " "
                    << olddir.X() << " " << olddir.Y() << " " << olddir.Z() << endl;
            }
            ph->SetWaveLength(lambda);  // [m]
            buf->Add(ph); 

         }
         i++;
      }      

      // update olddata and counter incriment //

      grandolddir = olddir;
      oldpos  = newpos;
      olddir  = newdir;

      fCurStepNo++;

      if (! fCurStepNo % 100) {
         cerr << fCurStepNo << " step finished!" << endl;
      }


   } // stepping loop end

}

//=====================================================================
//* Finish ------------------------------------------------------------

void IceRayTraceSim::Finish()
{
   fOutFile->cd("/photons");
   TIter next = fShellBuffers.GetIter();
   PhotonBuf *buf;
   while (buf = (PhotonBuf *)next()) {
      buf->Write();
   }
   fShellBuffers.Write();
   
   fOutFile->cd("/");
   fOutFile->Write();
   fOutFile->Close();
}

//=====================================================================
//* Draw ------------------------------------------------------------

void IceRayTraceSim::Draw()
{
   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "IceRayTraceSim::Draw called" << endl;
   }

   fHdNdL->SetLineWidth(4);
   fHdNdL->SetLineColor(kRed);
   fCanvas3 = new TCanvas("dNdLambda", "dNdLambda", 10, 10, 1200, 1200);
   fCanvas3->Divide(4, 4);
   fCanvas3->cd(1);
   fHdNdL->Draw();

   fCanvas1->cd(4);
   fHcostheta->SetLineWidth(4);
   fHcostheta->SetLineColor(kRed);
   fHcostheta->Draw();

   fCanvas4 = new TCanvas("incidentAng", "incidentAng", 10, 10, 1200, 1200);
   fCanvas4->Divide(4, 4);
   fCanvas5 = new TCanvas("incidentCos", "incidentCosTheta", 10, 10, 1200, 1200);
   fCanvas5->Divide(4, 4);
   fCanvas6 = new TCanvas("costheta_i", "costheta_i", 10, 10, 1200, 1200);
   fCanvas6->Divide(4, 4);
   fCanvas7 = new TCanvas("costheta_zi", "costheta_zi", 10, 10, 1200, 1200);
   fCanvas7->Divide(4, 4);
   fCanvas8 = new TCanvas("incidentCosz", "incidentCosTheta_z", 10, 10, 1200, 1200);
   fCanvas8->Divide(4, 4);
   fCanvas9 = new TCanvas("nsteps", "nsteps", 10, 10, 1200, 1200);
   fCanvas9->Divide(4, 4);
   fCanvas10 = new TCanvas("incidentAngz", "incidentAngle_z", 10, 10, 1200, 1200);
   fCanvas10->Divide(4, 4);

   for (Int_t i=0; i<fNshells; i++) {
      Int_t color = (i % 6) + 2;
      if (color == 5) {  // aboid yellow
         color = 28;
      }
      fHdNdL_d[i]->SetLineWidth(4);
      fHdNdL_d[i]->SetLineColor(color);
      fHinciAng_d[i]->SetMarkerStyle(6);
      fHinciAng_d[i]->SetMarkerSize(0.5);
      fHinciAng_d[i]->SetMarkerColor(color);
      fHinciCos_d[i]->SetMarkerStyle(6);
      fHinciCos_d[i]->SetMarkerSize(0.5);
      fHinciCos_d[i]->SetMarkerColor(color);
      fHinciAng_zd[i]->SetMarkerStyle(6);
      fHinciAng_zd[i]->SetMarkerSize(0.5);
      fHinciAng_zd[i]->SetMarkerColor(color);
      fHinciCos_zd[i]->SetMarkerStyle(6);
      fHinciCos_zd[i]->SetMarkerSize(0.5);
      fHinciCos_zd[i]->SetMarkerColor(color);
      fHcostheta_i[i]->SetLineWidth(4);
      fHcostheta_i[i]->SetLineColor(color);
      fHcostheta_zi[i]->SetLineWidth(4);
      fHcostheta_zi[i]->SetLineColor(color);
      fHnsteps_d[i]->SetLineWidth(4);
      fHnsteps_d[i]->SetLineColor(color);
      fCanvas3->cd(i+2); fHdNdL_d[i]->Draw();
      fCanvas4->cd(i+2); fHinciAng_d[i]->Draw();
      fCanvas5->cd(i+2); fHinciCos_d[i]->Draw();
      fCanvas6->cd(i+2); fHcostheta_i[i]->Draw();
      fCanvas7->cd(i+2); fHcostheta_zi[i]->Draw();
      fCanvas8->cd(i+2); fHinciAng_zd[i]->Draw();
      fCanvas9->cd(i+2); fHnsteps_d[i]->Draw();
      fCanvas10->cd(i+2); fHinciCos_zd[i]->Draw();
     
   }
}


//================================================================
// function GetCosTheta
//================================================================

Double_t IceRayTraceSim::GetCosTheta(const TVector3 &dir1,
                               const TVector3 &dir2)
{
   //
   // calculate costheta between dir1 and dir2.
   //

   Double_t ptot2 = dir1.Mag2() * dir2.Mag2();
   Double_t costheta;

   if(ptot2 <= 0) {
      cerr << " IceRayTraceSim::GetCosTheta: ptot2 <= 0! abort. " << endl;
      cerr << " dir1.Mag2, dir2.Mag2 = " << dir1.Mag2() << " " 
           << dir2.Mag2() << endl;
      abort();
   } else {
      costheta = dir1.Dot(dir2) / TMath::Sqrt(ptot2);
      if(costheta >  1.0) costheta =  1.0;
      if(costheta < -1.0) costheta = -1.0;
   }
   return costheta;
}

//================================================================
// function FindIntersection
//================================================================

Int_t IceRayTraceSim::FindIntersection(const TVector3 &oldpos, 
                                 const TVector3 &olddir,
                                 const TVector3 &newpos,
                                 Double_t curshell,
                                 TVector3 intersection[]) 
{
   //
   // find intersection of current shell and current step
   // Particle starts from oldpos with olddir and reache to newpos
   // Equations:
   // x = oldpos + t * olddir;
   // |x|^2 = curshell^2;
   // a*t^2 + 2*b*t + c = 0;
   //   where 
   //   a = |olddir|^2
   //   b = oldpos * olddir
   //   c = |oldpos|^2 - curshell^2
   //

   Double_t a = olddir.Mag2();
   Double_t b = oldpos * olddir;
   Double_t c = oldpos.Mag2() - curshell * curshell;

   // Is there any intersection?
   if ((b*b - a*c) < 0) {
      // no intersection!
      return 0;
   }

   Double_t d    = TMath::Sqrt(b*b - a*c);
   Double_t tpos = (-b + d) / a;
   Double_t tneg = (-b - d) / a;
   Double_t t[2];
   Int_t nfound  = 0;


   // 
   // Find intersection. 
   // Note that parameter t must be within a
   // range tmin < t <= tmax
   // (oldpos shouldn't be counted as a intersection of current step)
   //

   Double_t tmin = 0.;  // at oldpos
   Double_t tmax = (newpos - oldpos).Mag();  // at newpos

   if (d == 0.) { 
      if (tpos > tmin && tpos <= tmax) {  // tpos == tneg
         nfound = 1;
         t[0] = tpos;
      }
   } else { // d > 0
      if (tneg > tmin && tneg <= tmax) {
         nfound = 1;
         t[0] = tneg;
      }
      if (tpos > tmin && tpos <= tmax) {
         if (nfound) {
            t[1] = tpos;
         } else {
            t[0] = tpos;
         }
         nfound ++;
      }
   }


   for (Int_t i=0; i<nfound; i++) {
      intersection[i] = (oldpos + t[i]*olddir);
   }

   return nfound;

}

//================================================================
// function GetBeff
//================================================================

Double_t IceRayTraceSim::GetBeff(Double_t lambda) 
{ 
   //
   // Returns Beff coefficient at wavelength lambda[nm]
   //

   Double_t ratio = lambda / kRefLambda;
   Double_t beff  = kBeffAtRefLambda * TMath::Power(ratio, - kAlpha); // [m-1]

   return beff;

}

//================================================================
// function GetScatteringLength
//================================================================

Double_t IceRayTraceSim::GetScatteringLength(Double_t lambda) 
{ 
   //
   // Returns randomized scattering length[m].
   //

   Double_t random = gRandom->Uniform(0., 1.);
   Double_t lambdascatt   = 1. / GetBeff(lambda);
   Double_t scatteringlen = (- lambdascatt * TMath::Log(random)); // [m]

   //cerr << " lambda, lambdascatt, lograndom = " << lambda 
   //     << " " << lambdascatt << " " << TMath::Log(random) << endl;

   return scatteringlen;

}

//================================================================
// function DeviateVector
//================================================================

TVector3 IceRayTraceSim::DeviateVector(const TVector3 &orig, const TVector3 &dev) 
{ 
   // 
   // Return deviated direction vector (unit vector)
   // of "orig" with small deviation vector "dev".
   // Three vectors "axes[]" make referential frame
   // with the third component "axes[2]" along with "orig"
   // while others point arthogonal directions.
   // The "dev" vector must be defined in this referential frame.
   // It returns new vector:
   // dev.X() * axes[0] + dev.Y() * axes[1] + dev.Z() * axes[2].
   // 

   TVector3 glovalzaxis(0.,0.,1.);  // z-axis
   TVector3 axes[3];
   axes[2] = orig.Unit();

   if (fabs(axes[2].Z() - 1.) < 1.e-12) {
      axes[0].SetXYZ(1.,0.,0.);     // x-axis
   } else if (fabs(axes[2].Z() + 1.) < 1.e-12) {
      axes[0].SetXYZ(-1.,0.,0.);    // x-axis
   } else {
      axes[0] = (glovalzaxis.Cross(axes[2])).Unit();  // y-axis
   }

   axes[1] = (axes[2].Cross(axes[0])).Unit();
   TVector3 newdir(dev.X() * axes[0] + dev.Y() * axes[1] + dev.Z() * axes[2]);

#ifdef __DEBUG__

   cerr << " olddir = " << orig.X() << " " 
        << orig.Y() << " " << orig.Z() << endl;
   cerr << " newdir = " << newdir.X() << " " 
        << newdir.Y() << " " << newdir.Z() << endl;

#endif

   return newdir;

}

//================================================================
// function GetScatteringDirection
//================================================================

TVector3 IceRayTraceSim::GetScatteringDirection(const TVector3 &dir,
                                          TF1 *fDifCrossSecFunc)
{
   //    
   // Returns scattering direction of photon from original
   // direction &dir.
   // Set the differential cross section to the second argument
   // fDifCrossSecFunc.
   // 
   // define scattering direction from original "&dir"
   //

   Double_t costheta = TMath::Cos(fDifCrossSecFunc->GetRandom());
   Double_t sintheta = TMath::Sqrt((1 - costheta) * (1 + costheta));
   TVector3 dev(0, sintheta, costheta);  
   Double_t phi = gRandom->Uniform(0., 2*TMath::Pi());
   dev.RotateZ(phi);

   // 
   // return deviated vector
   //

   return DeviateVector(dir, dev);
 
}

//================================================================
// function GetAbsorptionLength
//================================================================

Double_t IceRayTraceSim::GetAbsorptionLength(Double_t lambda)
{ 

   //
   // Returns randomized absorption length[m].
   // 
   
   Double_t x[1], par[1], absorptioncoeff;
   x[0] = lambda; 
   
   absorptioncoeff = CalcAbsorptionCoefficient(x, par);

   Double_t lambdaabsorp    = 1. / absorptioncoeff;
   Double_t random          = gRandom->Uniform(0., 1.);
   
   Double_t absorptionlen   = (- lambdaabsorp * TMath::Log(random)); // [m]

   //cerr << " lambda, lambdaabsorp, lograndom = " << lambda
   //     << " " << lambdaabsorp << " " << TMath::Log(random) << endl;
   
   return absorptionlen;

}

//================================================================
// function Stepping
//================================================================

void IceRayTraceSim::Stepping(Double_t lambda, const TVector3 &oldpos, 
                        const TVector3 &olddir,
                        TF1 *fDsDcosthetaFunc,
                        TVector3 &newpos, TVector3 &newdir,
                        Bool_t &isabsorbed)
{
//
// Define one step.  
// Compete scattering process and absorption and
// choose smaller length as a winner.
//
// Arguments:
//  lambda ... wavelength of the traveling photon
//  oldpos ... start position of this step
//  olddir ... direction of stepping from oldpos
//  fDsDcosthetaFunc ... differential scattering angle cross section
//  newpos ... position of the end of the step
//  newdir ... direction of stepping from newpos
//  isabsorbed ... absorption flag
//
//  Note that the function must step a limited length.
//  If 0 length step is chosen, it repeats the step one more time.
//  
 
   newpos = oldpos;
   newdir = olddir;

   Double_t scatlen = GetScatteringLength(lambda);
   Double_t absolen = GetAbsorptionLength(lambda);

   if (scatlen == 0 || absolen == 0) {
      // Skip 0 length step. step again.
      scatlen = GetScatteringLength(lambda);
      absolen = GetAbsorptionLength(lambda);
   }

   //cerr << "lambda, scatlen, abslen = " 
   //     << lambda << "  " << scatlen << "  " << absolen << endl;

   if (scatlen < absolen) {

      isabsorbed = kFALSE;
      newpos = oldpos + scatlen * olddir;
      newdir = GetScatteringDirection(olddir, fDsDcosthetaFunc);      

   } else if (absolen < scatlen) {

      isabsorbed = kTRUE;
      newpos = oldpos + absolen * olddir;

   } else {

      cerr << "*************************************" << endl;
      cerr << "Absolen = Scatlen! called absorption." << endl;
      cerr << "*************************************" << endl;
      isabsorbed = kTRUE;
      newpos = oldpos + absolen * olddir;

   }

}

