//************************************************************************* 
//* --------------------
//* RayTrace
//* --------------------
//* (Description)
//*  Toy Monte-Carlo for photon propagation in Ice
//*  Main reference document is Ped's Ph-D thesis
//*  "Muon energy reconstruction in the Antarctic Muon And Neutrino Detector Array"
//* (Requires)
//* (Provides)
//* (Update Record)
//*	2004/04/15  K.Hoshina	Original version.
//*************************************************************************

#include "raytrace/RayTrace.h"
#include "raytrace/PhotonBuf.h"
#include "romeo/kern/RMOOpticalPhoton.h"
#include "romeo/kern/RMOUnit.h"
#include "TRandom.h"
#include "TMath.h"
#include <iostream>

ClassImp(RayTrace)

using namespace std;


const Double_t kGlambda   = 0.8;   // mean scattaring angle <costheta>
const Double_t kAlpha     = 0.84;  // expornential part of the effective scattering coefficient 
const Double_t kRefLambda = (532. * kNANO_METER);  // [m] wavelength of the caliblation YAG laser

//
// effective scattering coefficient for calibration light (532nm)
//
//const Double_t kBeffAtRefLambda = (1./24); // [m-1], averaged value
const Double_t kBeffAtRefLambda = 0.03;      // [m-1], at 1690m; clean ice
//const Double_t kBeffAtRefLambda = 0.05;    // [m-1], at 1740m; dust ice
//const Double_t kBeffAtRefLambda = 0.16;    // [m-1], at 2070m; most dusty ice

//
// absorption coefficient
//
const Double_t kA         = 8.0e+39;      // [m-1]; for UV cut off at L < 210nm
const Double_t kB         = 8300;         // [m-1]; for infrared absorption at L > 500nm
//const Double_t kCMdust    = (158 * kBeffAtRefLambda - 0.634); // 210nm < L < 500nm, averaged ice
const Double_t kCMdust    = 2.61;         // at 1690m, clean ice
//const Double_t kCMdust    = 7.12;       // at 1740m, dust ice
//const Double_t kCMdust    = 32.4;       // at 2040m, most dusty ice (calculated value)

const Double_t kKappa     = 1.07;           // m-1;

const Double_t kFineStrConst = (1./137);   // fine structure constant;

const Int_t kNxbinstheta = 2000;           // nbins of angle histograms
//const Int_t kNxbinstheta  = 20;         // nbins of angle histograms

const Double_t kOffsetShell = 100;        // m, offset radius of the shell sphere

//================================================================
// function DifferentialCrossSection
//================================================================

Double_t DifferentialCrossSection(Double_t *x, Double_t *par) 
{ 
   //
   // Calculate diffelential cross section of photon scattering
   // (used by TF1 function)
   //
   
   Double_t norm  = par[0];
   Double_t theta = x[0];

   Double_t grambda2 = kGlambda * kGlambda;
   Double_t denom    = 1 + grambda2 - 2 * kGlambda * TMath::Cos(theta);

   return (norm * (1 - grambda2) / TMath::Power(denom, 1.5) );

}

//================================================================
// function CerenkovWLSpectrum
//================================================================

Double_t CerenkovWLSpectrum(Double_t *x, Double_t *par) 
{ 

   Double_t norm = par[0];

   Double_t lambda = x[0] / kMICRO_METER;  // m to um

   Double_t n =   1.55749
                - 1.57988 * lambda
                + 3.99993 * lambda * lambda
                - 4.68271 * lambda * lambda * lambda
                + 2.09354 * lambda * lambda * lambda * lambda;

   Double_t beta = 1;                    // for relativistic muon
   Double_t costheta = 1 / (n * beta);
   Double_t sin2theta = (1 + costheta) * (1 - costheta);

   Double_t dNdL = 2 * TMath::Pi() * kFineStrConst * sin2theta
                   / (lambda * lambda);  

   return (norm * dNdL); 

}

//================================================================
// function CalcScatteringCoefficient
//================================================================

Double_t CalcScatteringCoefficient(Double_t *x, Double_t *par) 
{ 
   Double_t lambda = x[0];

   return RayTrace::GetBeff(lambda); 

}

//================================================================
// function CalcAbsorptionCoefficient
//================================================================

Double_t CalcAbsorptionCoefficient(Double_t *x, Double_t *par) 
{ 

   Double_t lambda = x[0] / kNANO_METER; // [m] to [nm]
   Double_t absorptioncoeff =   kA * TMath::Exp(-0.48175 * lambda) 
                              + kB * TMath::Exp(-6700 / lambda)
                              + kCMdust * TMath::Power(lambda, -kKappa); //

   return absorptioncoeff;

}

//================================================================
// function WLSpectrumAtR
//================================================================

Double_t WLSpectrumAtR(Double_t *x, Double_t *par) 
{ 

   Double_t lambda = x[0];
   Double_t dist   = par[0];
   Double_t norm   = par[2];
   Double_t dum[1];

   Double_t cwlpar[1];
   cwlpar[0] = norm; 

   Double_t origflax = CerenkovWLSpectrum(x, cwlpar);  
   Double_t scateff  = RayTrace::GetBeff(lambda);
   Double_t abseff   = CalcAbsorptionCoefficient(x, dum); 

   Double_t leffinv  = TMath::Sqrt(scateff * abseff);

   return (norm * origflax * TMath::Exp(- dist * leffinv));  

}

//=====================================================================
//* constructor -------------------------------------------------------

RayTrace::RayTrace(const char *name, const char *title)
       			: RMOVModule(name,title), fOutFile(0),
                          fDsDcosthetaFunc(0), fWLSpectrumFunc(0),
                          fAbsCoeff(0), fScatCoeff(0),
                          fCanvas1(0), fCanvas2(0), fCanvas3(0), fCanvas4(0), fCanvas5(0),
                          fHdNdL(0), fHdNdL_d(0), fWLSpect_r(0), fHinciAng_d(0),
                          fHinciCos_d(0)
{
   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "RayTrace:constructor is called" << endl;
   }

   fNshells    = 11;
   fRadiusStep = 5. * kMETER;        // m  // 100m ~ 150m 
   fLambdaMin  = 270 * kNANO_METER;  // m
   fLambdaMax  = 720 * kNANO_METER;  // m
   fOutFileName = "raytrace.root";
}

//=====================================================================
//* destructor --------------------------------------------------------

RayTrace::~RayTrace()
{

   if (fOutFile)         delete fOutFile;
   if (fDsDcosthetaFunc) delete fDsDcosthetaFunc;
   if (fWLSpectrumFunc)  delete fWLSpectrumFunc;
   if (fAbsCoeff)        delete fAbsCoeff;
   if (fScatCoeff)       delete fScatCoeff;
   if (fCanvas1)         delete  fCanvas1;
   if (fCanvas3)         delete  fCanvas3;
   if (fCanvas4)         delete  fCanvas4;
   if (fCanvas5)         delete  fCanvas5;
   if (fCanvas6)         delete  fCanvas6;
   if (fCanvas7)         delete  fCanvas7;
   if (fCanvas8)         delete  fCanvas8;
   if (fCanvas9)         delete  fCanvas9;
   if (fCanvas10)        delete  fCanvas10;

   // Are you sure you have set owner recursively?
   fShellBuffers.SetOwner();
   if (fShellBuffers.GetEntries()) fShellBuffers.RMOVAttBuffer::Delete();
}

//=====================================================================
//* Configure ---------------------------------------------------------

void RayTrace::Configure()
{
   //
   // Configure
   // Initializing following items:
   // - output rootfile
   // - Define functions 
   // - Draw functions in fCanvas1 
   // - Create Histograms
   //

   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "RayTrace::Configure is called!" << endl;
   }

   // 
   // create output rootfile
   //

   // save the original directory
   TDirectory *old_dir = gDirectory;

   // open file
   fOutFile = TFile::Open(fOutFileName, "recreate");

   //
   // create functions
   //

   fOutFile->cd("/");
   fOutFile->mkdir("functions", "functions");
   fOutFile->cd("/functions");

   fDsDcosthetaFunc = new TF1("DifferentialCrossSection", 
                               DifferentialCrossSection,
                               - TMath::Pi(), TMath::Pi(), 1);
   fDsDcosthetaFunc->SetLineWidth(4);
   fDsDcosthetaFunc->SetLineColor(kCyan);
   fDsDcosthetaFunc->SetNpx(5000);
   Double_t dcsfpar[1];
   dcsfpar[0] = 1.;                  // normalization factor
   fDsDcosthetaFunc->SetParameters(dcsfpar);
   Double_t integ = fDsDcosthetaFunc->Integral(- TMath::Pi(), TMath::Pi());
   dcsfpar[0] = 1 / integ;                  // normalization factor
   fDsDcosthetaFunc->SetParameters(dcsfpar);
   fDsDcosthetaFunc->Write();

   fWLSpectrumFunc = new TF1("CerenkovWLSpectrum", CerenkovWLSpectrum,
                                    fLambdaMin, fLambdaMax, 1);
   fWLSpectrumFunc->SetLineWidth(4);
   fWLSpectrumFunc->SetLineColor(kCyan);
   fWLSpectrumFunc->SetNpx(500);
   Double_t wlsfpar[1];
   //wlsfpar[0] = 1.;    // normalization factor
   wlsfpar[0] = 28000;    // normalization factor
   fWLSpectrumFunc->SetParameters(wlsfpar);
   fWLSpectrumFunc->Write();

   fAbsCoeff = new TF1("AbsCoeff", CalcAbsorptionCoefficient,
                                   fLambdaMin, fLambdaMax, 1);
   fAbsCoeff->SetParameters(dcsfpar);
   fAbsCoeff->SetLineWidth(4);
   fAbsCoeff->SetLineColor(kCyan);
   fAbsCoeff->SetNpx(500);
   fAbsCoeff->Write();

   fScatCoeff = new TF1("ScatLen", CalcScatteringCoefficient,
                                   fLambdaMin, fLambdaMax, 1);
   fScatCoeff->SetParameters(dcsfpar);
   fScatCoeff->SetLineWidth(4);
   fScatCoeff->SetLineColor(kCyan);
   fScatCoeff->SetNpx(500);
   fScatCoeff->Write();

   fCanvas1 = new TCanvas("Coeff", "Coeff", 10, 10, 800, 800);
   fCanvas1->Divide(2,2);
   fCanvas1->cd(1); fDsDcosthetaFunc->Draw();
   fCanvas1->cd(2); fScatCoeff->Draw();
   fCanvas1->cd(3); fAbsCoeff->Draw();
   fCanvas1->Write();

   if (GetVerboseLevel() == kPRECISE) {
      cerr << "functions are initialized!" << endl;
   }

   //
   // make histograms
   //

   fOutFile->cd("/");
   fOutFile->mkdir("hists", "hists");
   fOutFile->mkdir("photons", "photons");
   fOutFile->cd("/hists");

   Int_t nxbins = (Int_t)(fLambdaMax - fLambdaMin);
   fHdNdL       = new TH1D ("fHdNdL", "fHdNdL", nxbins, fLambdaMin, fLambdaMax); 
   fHdNdL->SetStats(1);

   fHdNdL_d      = new TH1D* [fNshells];
   fHinciAng_d   = new TH2D* [fNshells];
   fHinciCos_d   = new TH2D* [fNshells];
   fHinciAng_zd  = new TH2D* [fNshells];
   fHinciCos_zd  = new TH2D* [fNshells];
   fWLSpect_r    = new TF1*  [fNshells];
   fHcostheta_i  = new TH1D* [fNshells];
   fHcostheta_zi = new TH1D* [fNshells];
   fHnsteps_d    = new TH1D* [fNshells];
   fHcostheta    = new TH1D ("fHcostheta", "fHcostheta", 200, -1., 1.); 

   fHdNdL->SetXTitle("wave length [nm]");

   TCanvas *fCanvas2 = new TCanvas("WLSpectrum", "WLSpectrum", 10, 10, 1200, 1200);
   fCanvas2->Divide(4,4);
   fCanvas2->cd(1);
   fWLSpectrumFunc->Draw();

   for (Int_t i=0; i<fNshells; i++) {
      
      Int_t curR = fOffsetShellRadius + (Int_t)((i) * fRadiusStep);

      TString histname("fHdNdL_d");
      histname += curR;
      histname += "m";
      fHdNdL_d[i] = new TH1D(histname, histname, nxbins, fLambdaMin, fLambdaMax); 
      fHdNdL_d[i]->SetXTitle("wave length [nm]");
      fHdNdL_d[i]->SetStats(1);

      TString hist2name("fHinciAng_d");
      hist2name += curR;
      hist2name += "m";
      fHinciAng_d[i] = new TH2D(hist2name, hist2name, nxbins, fLambdaMin, fLambdaMax,
                                180, 0., 180.); 
      fHinciAng_d[i]->SetYTitle("angle of incidence (degree)");
      fHinciAng_d[i]->SetXTitle("wave length [nm]");
      fHinciAng_d[i]->SetStats(1);

      TString hist2zname("fHinciAng_zd");
      hist2zname += curR;
      hist2zname += "m";
      fHinciAng_zd[i] = new TH2D(hist2zname, hist2zname, kNxbinstheta, -1, 1,
                                kNxbinstheta, -1., 1.); 
      fHinciAng_zd[i]->SetYTitle("incidence angle from z-axis vs pos");
      fHinciAng_zd[i]->SetXTitle("pos (angle) ");
      fHinciAng_zd[i]->SetStats(1);

      TString hist3name("fHinciCos_d");
      hist3name += curR;
      hist3name += "m";
      fHinciCos_d[i] = new TH2D(hist3name, hist3name, nxbins, fLambdaMin, fLambdaMax,
                                kNxbinstheta, -1., 1.); 
      fHinciCos_d[i]->SetYTitle("incidence costheta");
      fHinciCos_d[i]->SetXTitle("wave length [nm]");
      fHinciCos_d[i]->SetStats(1);

      TString hist3zname("fHinciCos_zd");
      hist3zname += curR;
      hist3zname += "m";
      fHinciCos_zd[i] = new TH2D(hist3zname, hist3zname, 2000, -1, 1,
                                kNxbinstheta, -1., 1.); 
      fHinciCos_zd[i]->SetYTitle("incidence costheta from z-axis vs pos");
      fHinciCos_zd[i]->SetXTitle("pos (costheta) ");
      fHinciCos_zd[i]->SetStats(1);

      TString hist4name("fHcostheta_i");
      hist4name += i;
      fHcostheta_i[i] = new TH1D(hist4name, hist4name, kNxbinstheta, -1., 1.); 
      fHcostheta_i[i]->SetXTitle("costheta");
      fHcostheta_i[i]->SetStats(1);

      TString hist5name("fHcostheta_zi");
      hist5name += i;
      fHcostheta_zi[i] = new TH1D(hist5name, hist5name, kNxbinstheta, -1., 1.); 
      fHcostheta_zi[i]->SetXTitle("costheta from z-axis");
      fHcostheta_zi[i]->SetStats(1);

      TString hist6name("fHnsteps_d");
      hist6name += curR;
      hist6name += "m";
      fHnsteps_d[i] = new TH1D(hist6name, hist6name, 100, 0, 100);
      fHnsteps_d[i]->SetXTitle("number of steps");
      fHnsteps_d[i]->SetStats(1);

      TString funcname("WLSpecR");
      funcname += curR;
      funcname += "m";
      fWLSpect_r[i] = new TF1(funcname, WLSpectrumAtR, fLambdaMin, fLambdaMax, 3); 
      Double_t par[3];
      par[0] = (Double_t) curR;
      par[1] = 1;
      par[2] = 1;
      fWLSpect_r[i]->SetParameters(par);

      Int_t color = (i % 6) + 2;
      fWLSpect_r[i]->SetLineWidth(4);
      fWLSpect_r[i]->SetLineColor(color);
      fCanvas2->cd(i+2); fWLSpect_r[i]->Draw();

      TString bufname("PhotonBuf");
      bufname += curR;
      bufname += "m";
      PhotonBuf *buf = new PhotonBuf(bufname);
      cerr << "Photon buf created! name " << bufname << endl;
      buf->SetDistance(curR);
      // You are the owner! set owner now.
      buf->SetOwner();
      fShellBuffers.Add(buf);
   }
   // don't forget to set owner to fShellBuffers
   fShellBuffers.SetOwner();

   // back to the original directory
   old_dir->cd();
}

//=====================================================================
//* Process -----------------------------------------------------------

void RayTrace::Process()
{

   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "RayTrace::Process called" << endl;
   }

   Bool_t isfilled[fNshells];
   for (Int_t i=0; i<fNshells; i++) {
      isfilled[i] = kFALSE;
   }

   //Double_t lambda = fWLSpectrumFunc->GetRandom(fLambdaMin, fLambdaMax);
   Double_t lambda = 337.0 * kNANO_METER;  // [m],  flasher wave length


   fHdNdL->Fill(lambda);
   Double_t dcsfpar[1];
   dcsfpar[0] = lambda;
   fDsDcosthetaFunc->SetParameters(dcsfpar);

   TVector3 firstdir(1, 0, 0);
   TVector3 grandolddir(firstdir);
   TVector3 oldpos(0, 0, 0);
   TVector3 olddir(firstdir);
   TVector3 newpos(0, 0, 0);
   TVector3 newdir(0, 0, 0);
   Bool_t   isabsorbed = kFALSE;

   Double_t currentdist     = newpos.Mag();
   Double_t accumulatedist  = newpos.Mag();

   fCurStepNo = 0;
   
   //
   // stepping loop start
   //


   while (accumulatedist < 3 * (fOffsetShellRadius + (fNshells + 1) * fRadiusStep)) { 

      // excute one step //
      // olddir and newdir are unit vectors //

      Stepping(lambda, oldpos, olddir, fDsDcosthetaFunc,
               newpos, newdir, isabsorbed); 

      currentdist = newpos.Mag(); 
      Double_t currentsteplen = (newpos - oldpos).Mag();
      accumulatedist += currentsteplen;

      if (GetVerboseLevel() == kDEBUG) {
         cerr << "currentpos, steplen, currentdist = " 
              << newpos.X() << " " << newpos.Y() << " "
              << newpos.Z() << " " 
              << (newpos - oldpos).Mag() << " " 
              << currentdist << endl;
      }
      
      if (isabsorbed) {

         if (GetVerboseLevel() == kPRECISE) {
            cerr << "Photon is absorbed! pos, accumdist = "
                 << newpos.X() << " " << newpos.Y() << " "
                 << newpos.Z() << " " << accumulatedist << endl;
         }

         break;

      }

      if (fCurStepNo < fNshells) {
         Double_t curcostheta  = GetCosTheta(newpos, olddir);
         Double_t curcosthetaz = GetCosTheta(firstdir, olddir);
         fHcostheta_i[fCurStepNo]->Fill(curcostheta);
         fHcostheta_zi[fCurStepNo]->Fill(curcosthetaz);
      }

      // fill photonarray //

      Int_t i     = 0;
      TIter iter = fShellBuffers.GetIter();
      PhotonBuf *buf;
      while ((Int_t)(buf = (PhotonBuf *)iter.Next())) {
         //Double_t curshell = fOffsetShellRadius + i * fRadiusStep; 
         Double_t curshell = buf->GetDistance();
         Double_t olddist  = oldpos.Mag();
         Double_t newdist  = newpos.Mag();
         Int_t    nfoundx  = 0;
         TVector3 x[2];

         //
         //To save CPU time, calculate special cases first
         //
         if (olddist < curshell && newdist < curshell) {
            // it is impossible to intersect with current shell...
            nfoundx = 0; 

         } else if (olddist > curshell && (oldpos * olddir) > 0) {
            // photon fly away from current shell
            nfoundx = 0; 

         } else if (newdist == curshell) {
            // only one intersection 
            x[0] = newpos;
            nfoundx = 1;

         } else {
            // find intersection within the step range
            nfoundx = FindIntersection(oldpos, olddir, newpos, curshell, x);
         }

         for (Int_t k=0; k<nfoundx; k++) {

            RMOOpticalPhoton *ph = new RMOOpticalPhoton(x[k], olddir);
            if (GetVerboseLevel() == kDEBUG) {
               cerr << "dist, pos, dir = " << curshell << " " 
                    << x[k].X() << " " << x[k].Y() << " " << x[k].Z() << " "
                    << olddir.X() << " " << olddir.Y() << " " << olddir.Z() << endl;
            }
            ph->SetWaveLength(lambda);  // [m]
            buf->Add(ph); 

         }
         i++;
      }      

      // update olddata and counter incriment //

      grandolddir = olddir;
      oldpos  = newpos;
      olddir  = newdir;

      fCurStepNo++;

      if (! fCurStepNo % 100) {
         cerr << fCurStepNo << " step finished!" << endl;
      }


   } // stepping loop end

}

//=====================================================================
//* Finish ------------------------------------------------------------

void RayTrace::Finish()
{
   fOutFile->cd("/photons");
   TIter next = fShellBuffers.GetIter();
   PhotonBuf *buf;
   while (buf = (PhotonBuf *)next()) {
      buf->Write();
   }
   fShellBuffers.Write();
   
   fOutFile->cd("/");
   fOutFile->Write();
   fOutFile->Close();
}

//=====================================================================
//* Draw ------------------------------------------------------------

void RayTrace::Draw()
{
   if (GetVerboseLevel() == kFUNCTION_CALLED) {
      cerr << "RayTrace::Draw called" << endl;
   }

   fHdNdL->SetLineWidth(4);
   fHdNdL->SetLineColor(kRed);
   fCanvas3 = new TCanvas("dNdLambda", "dNdLambda", 10, 10, 1200, 1200);
   fCanvas3->Divide(4, 4);
   fCanvas3->cd(1);
   fHdNdL->Draw();

   fCanvas1->cd(4);
   fHcostheta->SetLineWidth(4);
   fHcostheta->SetLineColor(kRed);
   fHcostheta->Draw();

   fCanvas4 = new TCanvas("incidentAng", "incidentAng", 10, 10, 1200, 1200);
   fCanvas4->Divide(4, 4);
   fCanvas5 = new TCanvas("incidentCos", "incidentCosTheta", 10, 10, 1200, 1200);
   fCanvas5->Divide(4, 4);
   fCanvas6 = new TCanvas("costheta_i", "costheta_i", 10, 10, 1200, 1200);
   fCanvas6->Divide(4, 4);
   fCanvas7 = new TCanvas("costheta_zi", "costheta_zi", 10, 10, 1200, 1200);
   fCanvas7->Divide(4, 4);
   fCanvas8 = new TCanvas("incidentCosz", "incidentCosTheta_z", 10, 10, 1200, 1200);
   fCanvas8->Divide(4, 4);
   fCanvas9 = new TCanvas("nsteps", "nsteps", 10, 10, 1200, 1200);
   fCanvas9->Divide(4, 4);
   fCanvas10 = new TCanvas("incidentAngz", "incidentAngle_z", 10, 10, 1200, 1200);
   fCanvas10->Divide(4, 4);

   for (Int_t i=0; i<fNshells; i++) {
      Int_t color = (i % 6) + 2;
      if (color == 5) {  // aboid yellow
         color = 28;
      }
      fHdNdL_d[i]->SetLineWidth(4);
      fHdNdL_d[i]->SetLineColor(color);
      fHinciAng_d[i]->SetMarkerStyle(6);
      fHinciAng_d[i]->SetMarkerSize(0.5);
      fHinciAng_d[i]->SetMarkerColor(color);
      fHinciCos_d[i]->SetMarkerStyle(6);
      fHinciCos_d[i]->SetMarkerSize(0.5);
      fHinciCos_d[i]->SetMarkerColor(color);
      fHinciAng_zd[i]->SetMarkerStyle(6);
      fHinciAng_zd[i]->SetMarkerSize(0.5);
      fHinciAng_zd[i]->SetMarkerColor(color);
      fHinciCos_zd[i]->SetMarkerStyle(6);
      fHinciCos_zd[i]->SetMarkerSize(0.5);
      fHinciCos_zd[i]->SetMarkerColor(color);
      fHcostheta_i[i]->SetLineWidth(4);
      fHcostheta_i[i]->SetLineColor(color);
      fHcostheta_zi[i]->SetLineWidth(4);
      fHcostheta_zi[i]->SetLineColor(color);
      fHnsteps_d[i]->SetLineWidth(4);
      fHnsteps_d[i]->SetLineColor(color);
      fCanvas3->cd(i+2); fHdNdL_d[i]->Draw();
      fCanvas4->cd(i+2); fHinciAng_d[i]->Draw();
      fCanvas5->cd(i+2); fHinciCos_d[i]->Draw();
      fCanvas6->cd(i+2); fHcostheta_i[i]->Draw();
      fCanvas7->cd(i+2); fHcostheta_zi[i]->Draw();
      fCanvas8->cd(i+2); fHinciAng_zd[i]->Draw();
      fCanvas9->cd(i+2); fHnsteps_d[i]->Draw();
      fCanvas10->cd(i+2); fHinciCos_zd[i]->Draw();
     
   }
}


//================================================================
// function GetCosTheta
//================================================================

Double_t RayTrace::GetCosTheta(const TVector3 &dir1,
                               const TVector3 &dir2)
{
   //
   // calculate costheta between dir1 and dir2.
   //

   Double_t ptot2 = dir1.Mag2() * dir2.Mag2();
   Double_t costheta;

   if(ptot2 <= 0) {
      cerr << " RayTrace::GetCosTheta: ptot2 <= 0! abort. " << endl;
      cerr << " dir1.Mag2, dir2.Mag2 = " << dir1.Mag2() << " " 
           << dir2.Mag2() << endl;
      abort();
   } else {
      costheta = dir1.Dot(dir2) / TMath::Sqrt(ptot2);
      if(costheta >  1.0) costheta =  1.0;
      if(costheta < -1.0) costheta = -1.0;
   }
   return costheta;
}

//================================================================
// function FindIntersection
//================================================================

Int_t RayTrace::FindIntersection(const TVector3 &oldpos, 
                                 const TVector3 &olddir,
                                 const TVector3 &newpos,
                                 Double_t curshell,
                                 TVector3 intersection[]) 
{
   //
   // find intersection of current shell and current step
   // Particle starts from oldpos with olddir and reache to newpos
   // Equations:
   // x = oldpos + t * olddir;
   // |x|^2 = curshell^2;
   // a*t^2 + 2*b*t + c = 0;
   //   where 
   //   a = |olddir|^2
   //   b = oldpos * olddir
   //   c = |oldpos|^2 - curshell^2
   //

   Double_t a = olddir.Mag2();
   Double_t b = oldpos * olddir;
   Double_t c = oldpos.Mag2() - curshell * curshell;

   // Is there any intersection?
   if ((b*b - a*c) < 0) {
      // no intersection!
      return 0;
   }

   Double_t d    = TMath::Sqrt(b*b - a*c);
   Double_t tpos = (-b + d) / a;
   Double_t tneg = (-b - d) / a;
   Double_t t[2];
   Int_t nfound  = 0;


   // 
   // Find intersection. 
   // Note that parameter t must be within a
   // range tmin < t <= tmax
   // (oldpos shouldn't be counted as a intersection of current step)
   //

   Double_t tmin = 0.;  // at oldpos
   Double_t tmax = (newpos - oldpos).Mag();  // at newpos

   if (d == 0.) { 
      if (tpos > tmin && tpos <= tmax) {  // tpos == tneg
         nfound = 1;
         t[0] = tpos;
      }
   } else { // d > 0
      if (tneg > tmin && tneg <= tmax) {
         nfound = 1;
         t[0] = tneg;
      }
      if (tpos > tmin && tpos <= tmax) {
         if (nfound) {
            t[1] = tpos;
         } else {
            t[0] = tpos;
         }
         nfound ++;
      }
   }


   for (Int_t i=0; i<nfound; i++) {
      intersection[i] = (oldpos + t[i]*olddir);
   }

   return nfound;

}

//================================================================
// function GetBeff
//================================================================

Double_t RayTrace::GetBeff(Double_t lambda) 
{ 
   //
   // Returns Beff coefficient at wavelength lambda[nm]
   //

   Double_t ratio = lambda / kRefLambda;
   Double_t beff  = kBeffAtRefLambda * TMath::Power(ratio, - kAlpha); // [m-1]

   return beff;

}

//================================================================
// function GetScatteringLength
//================================================================

Double_t RayTrace::GetScatteringLength(Double_t lambda) 
{ 
   //
   // Returns randomized scattering length[m].
   //

   Double_t random = gRandom->Uniform(0., 1.);
   Double_t lambdascatt   = 1. / GetBeff(lambda);
   Double_t scatteringlen = (- lambdascatt * TMath::Log(random)); // [m]

   //cerr << " lambda, lambdascatt, lograndom = " << lambda 
   //     << " " << lambdascatt << " " << TMath::Log(random) << endl;

   return scatteringlen;

}

//================================================================
// function DeviateVector
//================================================================

TVector3 RayTrace::DeviateVector(const TVector3 &orig, const TVector3 &dev) 
{ 
   // 
   // Return deviated direction vector (unit vector)
   // of "orig" with small deviation vector "dev".
   // Three vectors "axes[]" make referential frame
   // with the third component "axes[2]" along with "orig"
   // while others point arthogonal directions.
   // The "dev" vector must be defined in this referential frame.
   // It returns new vector:
   // dev.X() * axes[0] + dev.Y() * axes[1] + dev.Z() * axes[2].
   // 

   TVector3 glovalzaxis(0.,0.,1.);  // z-axis
   TVector3 axes[3];
   axes[2] = orig.Unit();

   if (fabs(axes[2].Z() - 1.) < 1.e-12) {
      axes[0].SetXYZ(1.,0.,0.);     // x-axis
   } else if (fabs(axes[2].Z() + 1.) < 1.e-12) {
      axes[0].SetXYZ(-1.,0.,0.);    // x-axis
   } else {
      axes[0] = (glovalzaxis.Cross(axes[2])).Unit();  // y-axis
   }

   axes[1] = (axes[2].Cross(axes[0])).Unit();
   TVector3 newdir(dev.X() * axes[0] + dev.Y() * axes[1] + dev.Z() * axes[2]);

#ifdef __DEBUG__

   cerr << " olddir = " << orig.X() << " " 
        << orig.Y() << " " << orig.Z() << endl;
   cerr << " newdir = " << newdir.X() << " " 
        << newdir.Y() << " " << newdir.Z() << endl;

#endif

   return newdir;

}

//================================================================
// function GetScatteringDirection
//================================================================

TVector3 RayTrace::GetScatteringDirection(const TVector3 &dir,
                                          TF1 *fDifCrossSecFunc)
{
   //    
   // Returns scattering direction of photon from original
   // direction &dir.
   // Set the differential cross section to the second argument
   // fDifCrossSecFunc.
   // 
   // define scattering direction from original "&dir"
   //

   Double_t costheta = TMath::Cos(fDifCrossSecFunc->GetRandom());
   Double_t sintheta = TMath::Sqrt((1 - costheta) * (1 + costheta));
   TVector3 dev(0, sintheta, costheta);  
   Double_t phi = gRandom->Uniform(0., 2*TMath::Pi());
   dev.RotateZ(phi);

   // 
   // return deviated vector
   //

   return DeviateVector(dir, dev);
 
}

//================================================================
// function GetAbsorptionLength
//================================================================

Double_t RayTrace::GetAbsorptionLength(Double_t lambda)
{ 

   //
   // Returns randomized absorption length[m].
   // 
   
   Double_t x[1], par[1], absorptioncoeff;
   x[0] = lambda; 
   
   absorptioncoeff = CalcAbsorptionCoefficient(x, par);

   Double_t lambdaabsorp    = 1. / absorptioncoeff;
   Double_t random          = gRandom->Uniform(0., 1.);
   
   Double_t absorptionlen   = (- lambdaabsorp * TMath::Log(random)); // [m]

   //cerr << " lambda, lambdaabsorp, lograndom = " << lambda
   //     << " " << lambdaabsorp << " " << TMath::Log(random) << endl;
   
   return absorptionlen;

}

//================================================================
// function Stepping
//================================================================

void RayTrace::Stepping(Double_t lambda, const TVector3 &oldpos, 
                        const TVector3 &olddir,
                        TF1 *fDsDcosthetaFunc,
                        TVector3 &newpos, TVector3 &newdir,
                        Bool_t &isabsorbed)
{
//
// Define one step.  
// Compete scattering process and absorption and
// choose smaller length as a winner.
//
// Arguments:
//  lambda ... wavelength of the traveling photon
//  oldpos ... start position of this step
//  olddir ... direction of stepping from oldpos
//  fDsDcosthetaFunc ... differential scattering angle cross section
//  newpos ... position of the end of the step
//  newdir ... direction of stepping from newpos
//  isabsorbed ... absorption flag
//
//  Note that the function must step a limited length.
//  If 0 length step is chosen, it repeats the step one more time.
//  
 
   newpos = oldpos;
   newdir = olddir;

   Double_t scatlen = GetScatteringLength(lambda);
   Double_t absolen = GetAbsorptionLength(lambda);

   if (scatlen == 0 || absolen == 0) {
      // Skip 0 length step. step again.
      scatlen = GetScatteringLength(lambda);
      absolen = GetAbsorptionLength(lambda);
   }

   //cerr << "lambda, scatlen, abslen = " 
   //     << lambda << "  " << scatlen << "  " << absolen << endl;

   if (scatlen < absolen) {

      isabsorbed = kFALSE;
      newpos = oldpos + scatlen * olddir;
      newdir = GetScatteringDirection(olddir, fDsDcosthetaFunc);      

   } else if (absolen < scatlen) {

      isabsorbed = kTRUE;
      newpos = oldpos + absolen * olddir;

   } else {

      cerr << "*************************************" << endl;
      cerr << "Absolen = Scatlen! called absorption." << endl;
      cerr << "*************************************" << endl;
      isabsorbed = kTRUE;
      newpos = oldpos + absolen * olddir;

   }

}

