//************************************************************************* //* --------------------
//* RayTraceUtil
//* --------------------
//* (Description)
//*  Toy Monte-Carlo for photon propagation in Ice
//*  Main reference document is Ped's Ph-D thesis
//*  "Muon energy reconstruction in the Antarctic Muon And Neutrino Detector Array"
//* (Requires)
//* (Provides)
//* (Update Record)
//*	2004/04/15  K.Hoshina	Original version.
//*************************************************************************

#include "raytrace/RayTraceUtil.h"
#include "raytrace/PhotonBuf.h"
#include "raytrace/RayTracePhoton.h"
#include "romeo/kern/RMOUnit.h"
#include "TRandom.h"
#include "TMath.h"
#include <iostream>

ClassImp(RayTraceUtil)

using namespace std;


//================================================================
// class static members
//================================================================

RayTraceUtil * RayTraceUtil::fgCurrentUtil = 0;

//================================================================
// global parameters for TF1
//================================================================

static const Double_t kGlambda   = 0.8;   // mean scattaring angle <costheta>
static const Double_t kAlpha     = 0.84;  // expornential part of the effective scattering coefficient 
static const Double_t kRefLambda = (532. * kNANO_METER);  // [m] wavelength of the caliblation YAG laser
static const Double_t kKappa     = 1.07;                // m-1;
static const Double_t kFineStrConst = (1./137);         // fine structure constant;
static const Double_t kLambdaMin = (270 * kNANO_METER); // min wavelength [m]
static const Double_t kLambdaMax = (720 * kNANO_METER); // min wavelength [m]
static const Double_t kLightSpeed = 299792458;          // speed of light [m/s]


//
// absorption coefficient
//
static const Double_t kA         = 8.0e+39;      // [m-1]; for UV cut off at L < 210nm
static const Double_t kB         = 8300;         // [m-1]; for infrared absorption at L > 500nm


//================================================================
// global functions for TF1
//================================================================

//_______________________________________________________________
Double_t DifferentialCrossSection(Double_t *x, Double_t *par) 
{ 
//
// Calculate diffelential cross section of photon scattering
// (used by TF1 function)
//
   
   Double_t norm  = par[0];
   Double_t theta = x[0];

   Double_t grambda2 = kGlambda * kGlambda;
   Double_t denom    = 1 + grambda2 - 2 * kGlambda * TMath::Cos(theta);

   return (norm * (1 - grambda2) / TMath::Power(denom, 1.5) );

}

//_______________________________________________________________
Double_t GetPhaseIndex(Double_t *x, Double_t *) 
{
//
// Calculate phase index for specific wavelength.
// For details, see Ped's Ph-D thesis.
//
   Double_t lambda = x[0] / kMICRO_METER;  // m to um

   Double_t n =   1.55749
                - 1.57988 * lambda
                + 3.99993 * lambda * lambda
                - 4.68271 * lambda * lambda * lambda
                + 2.09354 * lambda * lambda * lambda * lambda;
   return n;
}

//_______________________________________________________________
Double_t GetGroupIndex(Double_t *x, Double_t *) 
{
//
// Calculate group index for specific wavelength.
// For details, see Ped's Ph-D thesis.
//
   Double_t lambda = x[0] / kMICRO_METER;  // m to um

   Double_t n =   1.55749
                - 3.99993 * lambda * lambda
                + 9.36542 * lambda * lambda * lambda
                - 6.28062 * lambda * lambda * lambda * lambda;
   return n;
}

//_______________________________________________________________
Double_t CherenkovWLSpectrum(Double_t *x, Double_t *par) 
{ 
//
// calculate cherenkov spectrum
//

   Double_t norm = par[0];
   Double_t lambda = x[0];

   Double_t n = GetPhaseIndex(x, par);

   Double_t beta = 1;                    // for relativistic muon
   Double_t costheta = 1 / (n * beta);
   Double_t sin2theta = (1 + costheta) * (1 - costheta);

   Double_t dNdL = 2 * TMath::Pi() * kFineStrConst * sin2theta
                   / (lambda * lambda);  

   return (norm * dNdL); 

}

//_______________________________________________________________
Double_t CalcScatteringCoefficient(Double_t *x, Double_t *) 
{ 
//
// 
//
   Double_t lambda = x[0];
   return RayTraceUtil::GetCurrentUtil()->GetBeff(lambda); 
}

//_______________________________________________________________
Double_t CalcAbsorptionCoefficient(Double_t *x, Double_t *) 
{ 

   Double_t lambda = x[0] / kNANO_METER; // [m] to [nm]
   Double_t cmdust = RayTraceUtil::GetCurrentUtil()->GetCMdust();
   Double_t absorptioncoeff =   kA * TMath::Exp(-0.48175 * lambda) 
                              + kB * TMath::Exp(-6700 / lambda)
                              + cmdust * TMath::Power(lambda, -kKappa);

   return absorptioncoeff;

}

//_______________________________________________________________
Double_t WaveLengthSpectrumAtR(Double_t *x, Double_t *par) 
{ 
//
// parameter x[0] is wavelength lambda
//
   Double_t dist   = par[0];
   Double_t norm   = par[1];
   Double_t dum[1];

   Double_t cwlpar[1];
   cwlpar[0] = norm; 

   Double_t origflax = CherenkovWLSpectrum(x, cwlpar);  
   Double_t scateff  = CalcScatteringCoefficient(x, dum);
   Double_t abseff   = CalcAbsorptionCoefficient(x, dum); 

   Double_t leffinv  = TMath::Sqrt(scateff * abseff);

   return (norm * origflax * TMath::Exp(- dist * leffinv));  

}

//================================================================
// Class functions 
//================================================================

//_______________________________________________________________
RayTraceUtil::RayTraceUtil()
       			: fDoInit(kTRUE),
                          fDsDcosthetaFunc(0), fCherenkovFunc(0),
                          fAbsCoeffFunc(0), fScatCoeffFunc(0),
                          fHdsdcostheta(0)
{
//
// Set effective scattering coefficient for calibration light(532nm)
// and absorption coefficient for 210nm < lambda < 500nm
//
// fBeffAtRefLambda = (1./24); // [m-1], averaged value
// fBeffAtRefLambda = 0.03;    // [m-1], at 1690m; clean ice
// fBeffAtRefLambda = 0.05;    // [m-1], at 1740m; dust ice
// fBeffAtRefLambda = 0.16;    // [m-1], at 2070m; most dusty ice
// fCMdust = (158 * kBeffAtRefLambda - 0.634); // averaged value
// fCMdust = 2.61; // at 1690m, clean ice
// fCMdust = 7.12; // at 1740m, dust ice
// fCMdust = 32.4; // at 2040m, most dusty ice(calculated value)
//

   fBeffAtRefLambda = 0.03;  // at 1690m, clean ice
   fCMdust = 2.61;           // at 1690m, clean ice
}

//_______________________________________________________________
RayTraceUtil::~RayTraceUtil()
{
   if (fDsDcosthetaFunc) delete fDsDcosthetaFunc;
   if (fCherenkovFunc)   delete fCherenkovFunc;
   if (fAbsCoeffFunc)    delete fAbsCoeffFunc;
   if (fScatCoeffFunc)   delete fScatCoeffFunc;
   if (fHdsdcostheta)    delete fHdsdcostheta;

   for (UInt_t i=0; i<fCanvases.size(); i++) {
      if (fCanvases[i]) delete fCanvases[i];
   }
}

//_______________________________________________________________
Double_t RayTraceUtil::GetLambdaMin()
{
   return kLambdaMin;
}

//_______________________________________________________________
Double_t RayTraceUtil::GetLambdaMax()
{
   return kLambdaMax;
}

//_______________________________________________________________
Double_t RayTraceUtil::GetBeff(Double_t lambda) 
{ 
//
// Returns Beff coefficient at wavelength lambda[nm]
//

   Double_t ratio = lambda / kRefLambda;
   Double_t beff  = fBeffAtRefLambda * TMath::Power(ratio, - kAlpha); // [m-1]

   return beff;

}

//_______________________________________________________________
void RayTraceUtil::InitFunc()
{ 
//
// Call SetCurrentUtil to specify current active utility.
// initialize functions if fDoInit flag is kTRUE
// available functions:
//
// - fDsDcosthetaFunc
//   Differential cross section of photon scattering.
//   Integral over -pi to pi is 1.
//   valiable -- theta [rad] (scattering angle, [-pi, pi])
//   param[0] -- normalization factor
//
// - fCherenkovFunc
//   Cherenkov wavelength spectrum from kLambdaMin to kLambdaMax. 
//   virtical axis is arbitrary.
//   valiable -- lambda [m] (wavelength, [kLambdaMin < lambda < kLambdaMax])
//   param[0] -- scale factor 
//
// - fAbsCoeffFunc
//   absorption coefficient as a function of wavelength.
//   no parameter is required
//   valiable -- lambda [m] (wavelength, [kLambdaMin < lambda < kLambdaMax])
//    
// - fScatCoeffFunc
//   scattering coefficient as a function of wavelength.
//   no parameter is required
//   valiable -- lambda [m] (wavelength, [kLambdaMin < lambda < kLambdaMax])
//

   SetCurrentUtil(this);
   if (fDoInit == kFALSE) return;

   // scattering angle
   if (!fDsDcosthetaFunc) {

      fDsDcosthetaFunc = new TF1("DifferentialCrossSection",
                                  DifferentialCrossSection,
                                  - TMath::Pi(), TMath::Pi(), 1);
      fDsDcosthetaFunc->SetLineWidth(4);
      fDsDcosthetaFunc->SetLineColor(kCyan);
      fDsDcosthetaFunc->SetNpx(5000);
      Double_t dcsfpar[1];
      dcsfpar[0] = 1.;  
      fDsDcosthetaFunc->SetParameters(dcsfpar);
      Double_t integ = fDsDcosthetaFunc->Integral(- TMath::Pi(), TMath::Pi());
      dcsfpar[0] = 1 / integ;   // normalization factor
      fDsDcosthetaFunc->SetParameters(dcsfpar);
      //fDsDcosthetaFunc->GetXaxis()->SetTitle("scattering theta [rad]");
      //fDsDcosthetaFunc->GetYaxis()->SetTitle("differential scattering cross section");
   }

   // cherenkov spectrum
   if (!fCherenkovFunc) {
      fCherenkovFunc = new TF1("CherenkovWLSpectrum", CherenkovWLSpectrum,
                                 kLambdaMin, kLambdaMax, 1);
      fCherenkovFunc->SetLineWidth(4);
      fCherenkovFunc->SetLineColor(kCyan);
      fCherenkovFunc->SetNpx(500);
      Double_t wlsfpar[1];
      wlsfpar[0] = 1.;    // scale factor, arbitrary
      fCherenkovFunc->SetParameters(wlsfpar);
      //fCherenkovFunc->GetXaxis()->SetTitle("wavelength[m]");
      //fCherenkovFunc->GetYaxis()->SetTitle("Cherenkov spectrum");
   }


   if (fAbsCoeffFunc)    delete fAbsCoeffFunc;
   if (fScatCoeffFunc)   delete fScatCoeffFunc;
   if (fHdsdcostheta)    delete fHdsdcostheta;


   fAbsCoeffFunc = new TF1("AbsoptionCoefficient", CalcAbsorptionCoefficient,
                            kLambdaMin, kLambdaMax, 1); 
                              // no parameter needed
   fAbsCoeffFunc->SetLineWidth(4);
   fAbsCoeffFunc->SetLineColor(kCyan);
   fAbsCoeffFunc->SetNpx(500);
   //fAbsCoeffFunc->GetXaxis()->SetTitle("wavelength[m]");
   //fAbsCoeffFunc->GetYaxis()->SetTitle("absorption coefficient");

   fScatCoeffFunc = new TF1("ScatteringCoefficient", CalcScatteringCoefficient,
                           kLambdaMin, kLambdaMax, 1); 
   fScatCoeffFunc->SetLineWidth(4);
   fScatCoeffFunc->SetLineColor(kCyan);
   fScatCoeffFunc->SetNpx(500);
   //fScatCoeffFunc->GetXaxis()->SetTitle("wavelength[m]");
   //fScatCoeffFunc->GetYaxis()->SetTitle("scattering coefficient");

   // histgrams for debugging purpose...
   fHdsdcostheta = new TH1D("hDsDcostheta","hist of differential cross section", 
                            360, -TMath::Pi(), TMath::Pi()); 

   fDoInit = kFALSE;
}


//_______________________________________________________________
TF1* RayTraceUtil::CreateCherenkovSpectrumAtDistFunc(Double_t dist)
{
   TString title("WaveLengthSpectrumAtR"); 
   title += (Int_t)dist;
   TF1* func = new TF1(title, WaveLengthSpectrumAtR,
                      kLambdaMin, kLambdaMax, 2);
   func->SetLineWidth(4);
   func->SetLineColor(kCyan);
   func->SetNpx(500);
   Double_t par[2];
   par[0] = dist;    // distance from origin
   par[1] = 1 ;      // scale factor, arbitrary
   func->SetParameters(par);

   return func;
}

//_______________________________________________________________
void RayTraceUtil::Draw()
{
   TCanvas *c1 = new TCanvas("functions","functions",10, 10, 800, 800);
   fCanvases.push_back(c1);
   c1->Divide(2,2);

   c1->cd(1); const_cast<TF1&>(GetDiffCrossSectionFunc()).Draw();
   c1->cd(2); const_cast<TF1&>(GetCherenkovSpectrumFunc()).Draw();
   c1->cd(3); const_cast<TF1&>(GetAbsorptionCoeffFunc()).Draw();
   c1->cd(4); const_cast<TF1&>(GetScatterCoeffFunc()).Draw();
   c1->Draw();

   TCanvas *c2 = new TCanvas("Histos","Histos",30, 30, 400, 400);
   fCanvases.push_back(c2);

   c2->cd(); fHdsdcostheta->Draw();

}

//_______________________________________________________________
Double_t RayTraceUtil::Stepping(Double_t lambda, 
                                const TVector3 &oldpos, 
                                const TVector3 &olddir,
                                const TF1 &funcDsDcostheta,
                                TVector3 &newpos, 
                                TVector3 &newdir,
                                Double_t &scatcostheta,
                                Bool_t &isabsorbed)
{
//
// Execute one step and returns time[sec] consumption during the step.
// Compete scattering process and absorption and
// choose smaller length as a winner.
//
// Arguments:
//  lambda ... wavelength of the traveling photon
//  oldpos ... start position of this step
//  olddir ... direction of stepping from oldpos
//  funcDsDcostheta ... differential scattering angle cross section
//  newpos ... position of the end of the step
//  newdir ... direction of stepping from newpos
//  isabsorbed ... absorption flag
//
//  Note that the function must step a limited length.
//  If 0 length step is chosen, it repeats the step one more time.
//  
 
   newpos = oldpos;
   newdir = olddir;

   Double_t scatlen = GetScatteringLength(lambda);
   Double_t absolen = GetAbsorptionLength(lambda);
   Double_t winner = 0;

   if (scatlen == 0 || absolen == 0) {
      // Skip 0 length step. step again.
      scatlen = GetScatteringLength(lambda);
      absolen = GetAbsorptionLength(lambda);
   }

   //cerr << "lambda, scatlen, abslen = " 
   //     << lambda << "  " << scatlen << "  " << absolen << endl;

   if (scatlen < absolen) {  // photon is scatterd!

      isabsorbed = kFALSE;
      winner = scatlen;
      newdir = GetScatteringDirection(olddir, funcDsDcostheta, scatcostheta);      

   } else if (absolen < scatlen) { // photon is absorbed!

      isabsorbed = kTRUE;
      winner = absolen;

   } else {

      cerr << "*************************************" << endl;
      cerr << "Absolen = Scatlen! called absorption." << endl;
      cerr << "*************************************" << endl;
      isabsorbed = kTRUE;
      winner = absolen;

   }
   newpos = oldpos + winner * olddir;

   Double_t par[1];
   par[0] = lambda; 
   Double_t icevelocity = kLightSpeed / GetGroupIndex(par, par);
   Double_t time = winner / icevelocity; 
  
   return time;  // in [sec]

}


//_______________________________________________________________
Double_t RayTraceUtil::GetScatteringLength(Double_t lambda) 
{ 
   //
   // Returns randomized scattering length[m].
   //

   Double_t random = gRandom->Uniform(0., 1.);
   Double_t lambdascatt   = 1. / GetBeff(lambda);
   Double_t scatteringlen = (- lambdascatt * TMath::Log(random)); // [m]

   //cerr << " lambda, lambdascatt, lograndom = " << lambda 
   //     << " " << lambdascatt << " " << TMath::Log(random) << endl;

   return scatteringlen;

}

//_______________________________________________________________
TVector3 RayTraceUtil::DeviateVector(const TVector3 &orig, const TVector3 &dev) 
{ 
// 
// Return deviated direction vector (unit vector)
// of "orig" with a deviation vector "dev".
// Three vectors "axes[]" make referential frame
// with the third component "axes[2](local z-axis)" along with "orig" vector.
// The "dev" vector must be defined in this referential frame.
// It returns new vector:
// dev.X() * axes[0] + dev.Y() * axes[1] + dev.Z() * axes[2].
// 

   TVector3 glovalzaxis(0.,0.,1.);  // global z-axis
   TVector3 axes[3];
   axes[2] = orig.Unit(); // local z-axis

   if (fabs(axes[2].Z() - 1.) < 1.e-12) {
      // almost parallel to global z-axis! set local x-axis as global one.
      axes[0].SetXYZ(1.,0.,0.);     // local x-axis

   } else if (fabs(axes[2].Z() + 1.) < 1.e-12) {
      // almost oposite direction to global z-axis! 
      // set local x-axis as oposite global x-axis.
      axes[0].SetXYZ(-1.,0.,0.);    // local x-axis

   } else {
      axes[0] = (glovalzaxis.Cross(axes[2])).Unit();  // local x-axis
   }

   axes[1] = (axes[2].Cross(axes[0])).Unit(); // local y-axis
   TVector3 newdir(dev.X() * axes[0] + dev.Y() * axes[1] + dev.Z() * axes[2]);

#ifdef __DEBUG__

   cerr << " olddir = " << orig.X() << " " 
        << orig.Y() << " " << orig.Z() << endl;
   cerr << " newdir = " << newdir.X() << " " 
        << newdir.Y() << " " << newdir.Z() << endl;

#endif

   return newdir;

}

//================================================================
// function GetScatteringDirection
//================================================================

TVector3 RayTraceUtil::GetScatteringDirection(const TVector3 &dir,
                                              const TF1 &funcDiffCrossSection,
                                              Double_t &scatcostheta)
{
//    
// Returns scattering direction of photon from original
// direction &dir.
// Set the differential cross section to the second argument
// funcDiffCrossSection.
// 
// It creates deviation vector (dev) and give it to the DeviateVector
// function. This deviation vector is defined in local coordinate,
// which z-axis is exactly aligned with the original direction &dir.
// Therefore dev(0, sintheta, costheta) indicates a vector in the y-z plane
// of the local coordinate, with zenith angle theta.
// Then the dev vector is rotated around z-axis with the random angle.
//

   Double_t theta = (const_cast<TF1&>(funcDiffCrossSection)).GetRandom();
   fHdsdcostheta->Fill(theta);
   scatcostheta = TMath::Cos(theta);
   //fHdsdcostheta->Fill(scatcostheta);
   Double_t sintheta = TMath::Sqrt((1 - scatcostheta) * (1 + scatcostheta));
   TVector3 dev(0, sintheta, scatcostheta);  
   Double_t phi = gRandom->Uniform(0., 2*TMath::Pi());
   dev.RotateZ(phi);

   // 
   // return deviated vector
   //

   return DeviateVector(dir, dev);
 
}

//================================================================
// function GetAbsorptionLength
//================================================================

Double_t RayTraceUtil::GetAbsorptionLength(Double_t lambda)
{ 

   //
   // Returns randomized absorption length[m].
   // 
   
   Double_t x[1], par[1], absorptioncoeff;
   x[0] = lambda; 
   
   absorptioncoeff = CalcAbsorptionCoefficient(x, par);

   Double_t lambdaabsorp    = 1. / absorptioncoeff;
   Double_t random          = gRandom->Uniform(0., 1.);
   
   Double_t absorptionlen   = (- lambdaabsorp * TMath::Log(random)); // [m]

   //cerr << " lambda, lambdaabsorp, lograndom = " << lambda
   //     << " " << lambdaabsorp << " " << TMath::Log(random) << endl;
   
   return absorptionlen;

}

#if 0
//_______________________________________________________________
Double_t RayTraceUtil::GetCosTheta(const TVector3 &dir1,
                                   const TVector3 &dir2)
{
   //
   // calculate costheta between dir1 and dir2.
   //

   Double_t ptot2 = dir1.Mag2() * dir2.Mag2();
   Double_t costheta;

   if(ptot2 <= 0) {
      cerr << " RayTraceUtil::GetCosTheta: ptot2 <= 0! abort. " << endl;
      cerr << " dir1.Mag2, dir2.Mag2 = " << dir1.Mag2() << " " 
           << dir2.Mag2() << endl;
      abort();
   } else {
      costheta = dir1.Dot(dir2) / TMath::Sqrt(ptot2);
      if(costheta >  1.0) costheta =  1.0;
      if(costheta < -1.0) costheta = -1.0;
   }
   return costheta;
}
#endif
