#ifndef __RayTraceUtil__H__
#define __RayTraceUtil__H__

///////////////////////////////////////////////////////////////////////////
/// ---------------------
/// RayTraceUtil Class
/// ---------------------
/// (Description)
///     utility class for ray tracer. Provides photon spectrum such as
///     Cherenkov, Scattering & Absorption coefficient for given wavelength
///     and ice properties.
///     It also provides stepping function of traced photon.
/// (Update Record)
///	2006/01/20  K.Hoshina	Original version.
///////////////////////////////////////////////////////////////////////////

#include "TVector3.h"
#include "TH1D.h"
#include "TF1.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TRandom.h"
#include "TObjArray.h"
#include <vector>

using namespace std;

class RayTraceUtil {

public:

  RayTraceUtil();
  virtual ~RayTraceUtil();

  /**
   * return TF1 function of differential cross section 
   */
  virtual const TF1& GetDiffCrossSectionFunc()  
                                { InitFunc(); return *fDsDcosthetaFunc; }
  //----------------------------------------------------------------------

  /**
   * return function of Cherenkov spectrum
   */
  virtual const TF1& GetCherenkovSpectrumFunc() 
                                { InitFunc(); return *fCherenkovFunc; }
  //----------------------------------------------------------------------

  /**
   * return function of Absorption Coefficient vs wavelength 
   */
  virtual const TF1& GetAbsorptionCoeffFunc() 
                                { InitFunc(); return *fAbsCoeffFunc; } 
  //----------------------------------------------------------------------

  /**
   * return function of Scattering Length vs wavelength
   */
  virtual const TF1& GetScatterCoeffFunc()
                                { InitFunc(); return *fScatCoeffFunc; } 
  //----------------------------------------------------------------------

  /**
   * return function of Cherenkov spectrum after dist[m] propagation.
   * Using Standard pole ice model.
   * It creates new TF1 so that users must delete it.
   */
  virtual TF1* CreateCherenkovSpectrumAtDistFunc(Double_t dist);
  //----------------------------------------------------------------------

  /**
   * return range of wavelength
   */
  static Double_t GetLambdaMin();
  static Double_t GetLambdaMax();
  //----------------------------------------------------------------------
  
  /**
   * set/get effective scattering coefficient(Beff) @ lambda = kRefLambda
   */
  inline void SetBeffAtRefLambda(Double_t x) 
                                { fBeffAtRefLambda = x; fDoInit = kTRUE; }
  virtual Double_t GetBeffAtRefLambda() const { return fBeffAtRefLambda; }
  //----------------------------------------------------------------------

  /**
   * set absorption coefficient from 210nm < lambda < 500nm
   */
  inline void SetCMdust(Double_t x) { fCMdust  = x; fDoInit = kTRUE; }
  virtual Double_t GetCMdust() const { return fCMdust; }
  //----------------------------------------------------------------------

  /**
   * Get pointer to current util
   */
  static RayTraceUtil* GetCurrentUtil() { return fgCurrentUtil; }
  static void  SetCurrentUtil(RayTraceUtil* me) { fgCurrentUtil = me; }
  //----------------------------------------------------------------------

  /**
   * get effective scattering coefficient(Beff) at given lambda
   */
  virtual Double_t GetBeff(Double_t lambda);
  //----------------------------------------------------------------------

  /**
   * Draw functions and histgrams
   */
  virtual void Draw();
  //----------------------------------------------------------------------

protected:

  virtual void     InitFunc();
  virtual Double_t Stepping(Double_t lambda, 
                            const TVector3 &oldpos, 
                            const TVector3 &olddir,
                            const TF1 &dcsf,
                                  TVector3 &newpos, 
                                  TVector3 &newdir,
                                  Double_t &scatcostheta,
                                  Bool_t &isabsorbed);

private:

  // following functions are used inside the Stepping function

  Double_t GetScatteringLength(Double_t lambda);
  Double_t GetAbsorptionLength(Double_t lambda);
  TVector3 GetScatteringDirection(const TVector3 &dir, const TF1 &dcsf, 
                                  Double_t &scatcostheta);
  TVector3 DeviateVector(const TVector3 &orig, const TVector3 &dev);

private:

  Bool_t   fDoInit;          // flag to excute InitFunc
  Double_t fBeffAtRefLambda; // effective scattering coefficient @532nm
  Double_t fCMdust;          // absorption coefficient at 210nm < lambda < 500nm 

  TF1     *fDsDcosthetaFunc; // function of differential cross section 
  TF1     *fCherenkovFunc;   // function of wave length spectrum ov cerenkov
  TF1     *fAbsCoeffFunc;    // function of Absorption Coefficient 
  TF1     *fScatCoeffFunc;   // function of Scattering Length

  TH1D    *fHdsdcostheta;    // hist of differential cross section 

  vector<TCanvas*> fCanvases; // vector of canvas

  static RayTraceUtil* fgCurrentUtil; // pointer to current RayTraceUtil

  ClassDef(RayTraceUtil, 1)       // Ray trace Monte-Carlo program

};

#endif

